# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

import enum
import logging

from gitfourchette import colors
from gitfourchette.forms.ui_searchbar import Ui_SearchBar
from gitfourchette.globalshortcuts import GlobalShortcuts
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.search.searchprovider import SearchProvider
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class SearchBar(QWidget):
    class Op(enum.IntEnum):
        Start = enum.auto()
        Next = enum.auto()
        Previous = enum.auto()

    DebounceDelayMs = 250
    MacToolTipQuirks = MACOS
    _ToolTipTag = "<!--SearchBarToolTip-->"

    buddy: QWidget
    """
    Widget in which the search is carried out.
    """

    provider: SearchProvider | None
    """
    Current search provider.
    Receives the search term as input, performs the actual search.
    """

    providers: tuple[SearchProvider, ...]
    """
    All search providers supported by this search bar.
    """

    debounceTimer: QTimer
    """
    Schedules the SearchProvider to update the search results after the user
    stops typing.
    """

    autoJumpWhenResultsComeIn: bool
    """
    If True, debouncing will jump to an occurrence of the search term, if found.
    Otherwise, the search term will be reevaluated without changing the
    selection. This flag is reset to True whenever a debounce is rescheduled.
    """

    @property
    def rawSearchTerm(self) -> str:
        return self.lineEdit.text()

    @property
    def lineEdit(self) -> QLineEdit:
        return self.ui.lineEdit

    @property
    def buttons(self):
        return (self.ui.providerChooser,
                self.ui.forwardButton,
                self.ui.backwardButton,
                self.ui.closeButton,
                self.ui.filterCheckBox)

    def __init__(self, buddy: QWidget, *providers: SearchProvider):
        super().__init__(buddy)

        self.setObjectName(f"SearchBar({buddy.objectName()})")
        self.buddy = buddy

        self.ui = Ui_SearchBar()
        self.ui.setupUi(self)

        self.lineEdit.addAction(stockIcon("magnifying-glass"), QLineEdit.ActionPosition.LeadingPosition)
        self.loupe: QAction = self.lineEdit.actions()[0]

        self.lineEdit.textChanged.connect(self.onSearchTextChanged)
        self.ui.filterCheckBox.toggled.connect(self.onFilterCheckBoxToggled)

        self.ui.closeButton.clicked.connect(self.bail)
        self.ui.forwardButton.clicked.connect(self.runSearchForward)
        self.ui.backwardButton.clicked.connect(self.runSearchBackward)

        self.ui.forwardButton.setIcon(stockIcon("go-down-search"))
        self.ui.backwardButton.setIcon(stockIcon("go-up-search"))
        self.ui.closeButton.setIcon(stockIcon("dialog-close"))

        self._providerChooserMenu = QMenu(self)
        ActionDef.addSectionToQMenu(self._providerChooserMenu, _p("SearchBar", "Search scope"))
        self._providerChooserActionGroup = QActionGroup(self)
        self.ui.providerChooser.setMenu(self._providerChooserMenu)
        self.ui.providerChooser.setVisible(False)
        self.ui.providerChooser.setFixedWidth(80)

        # The size of the buttons is readjusted after show(),
        # so prevent visible popping when booting up for the first time.
        for button in self.buttons:
            button.setMaximumHeight(1)

        appendShortcutToToolTip(self.ui.backwardButton, GlobalShortcuts.findPrevious[0])
        appendShortcutToToolTip(self.ui.forwardButton, GlobalShortcuts.findNext[0])
        appendShortcutToToolTip(self.ui.closeButton, Qt.Key.Key_Escape)

        self.debounceTimer = QTimer(self)
        self.debounceTimer.setSingleShot(True)
        self.debounceTimer.setInterval(self.DebounceDelayMs if not APP_TESTMODE else 0)
        self.debounceTimer.timeout.connect(self.onDebounce)

        self.autoJumpWhenResultsComeIn = True

        tweakWidgetFont(self.ui.lineEdit, 85)
        tweakWidgetFont(self.ui.filterCheckBox, 85)
        tweakWidgetFont(self.ui.providerChooser, 85)

        withChildren = Qt.ShortcutContext.WidgetWithChildrenShortcut
        makeWidgetShortcut(self, self.onEnterShortcut, "Return", "Enter", context=withChildren)
        makeWidgetShortcut(self, self.onShiftEnterShortcut, "Shift+Return", "Shift+Enter", context=withChildren)
        makeWidgetShortcut(self, self.bail, "Escape", context=withChildren)

        if SearchBar.MacToolTipQuirks:
            self.trapNextToolTip = False
            self.trappedToolTip = ""

        assert providers, "SearchBar needs at least one SearchProvider"
        self.provider = None
        self.providers = providers
        for provider in providers:
            self._installProvider(provider)
        self.ui.providerChooser.setVisible(len(providers) >= 2)
        self._setProvider(self.providers[0])

    def _installProvider(self, provider: SearchProvider):
        provider.statusChanged.connect(self.onProviderStatusChanged)
        provider.freeze(True)  # Freeze until in active use

        def makeCurrent():
            self._setProvider(provider)

        action = QAction(f"{provider.shortTitle()} \u2013 {provider.longTitle()}", parent=self)
        action.setActionGroup(self._providerChooserActionGroup)
        action.setCheckable(True)
        action.setData(provider)
        action.triggered.connect(makeCurrent)

        keys = provider.keyboardShortcut()
        if keys:
            action.setShortcut(keys)

            # Prevent action from being invoked by shortcut application-wide,
            # and duplicate the shortcut so it works throughout the SearchBar
            action.setShortcutContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
            makeWidgetShortcut(self, makeCurrent, keys, context=Qt.ShortcutContext.WidgetWithChildrenShortcut)

        self._providerChooserMenu.addAction(action)

    def onEnterShortcut(self):
        if SearchBar.MacToolTipQuirks:
            self.trapNextToolTip = True

        self.lineEdit.selectAll()
        self.runSearchForward()

    def onShiftEnterShortcut(self):
        if SearchBar.MacToolTipQuirks:
            self.trapNextToolTip = True

        self.lineEdit.selectAll()
        self.runSearchBackward()

    def runSearchForward(self):
        self.runSearch(True)

    def runSearchBackward(self):
        self.runSearch(False)

    def runSearch(self, forward: bool):
        wasAwaitingDebounce = self.debounceTimer.isActive()

        self.debounceTimer.stop()
        self.hideToolTip()

        # We're here because the user has hit the return key, and...
        # ...the query is still running, so tell user to chill
        if self.provider.status() == SearchProvider.TermStatus.Loading:
            self.showToolTip(_("Please wait…"))

        # ...user is in a hurry, expedite debounce
        elif wasAwaitingDebounce:
            self.onDebounce(forward)

        # ...we know there are matches, so jump to next match
        elif self.provider.isGoodAndNonEmpty():
            self.provider.jump(forward)

        # ...the search term is empty, so quack at user
        elif self.provider.isEmpty():
            QApplication.beep()

        # ...there are no results, so tell user about it explicitly
        elif self.provider.isBad():
            self.showToolTip(self.provider.notFoundMessage())

    def onDebounce(self, forwardHint: bool = True):
        assert self.isVisible(), "don't debounce while invisible"
        assert self.provider.status() != SearchProvider.TermStatus.Loading
        if self.provider.isEmpty() or self.provider.isBad():
            return

        self.hideToolTip()

        # Prime the search. This may emit provider.statusChanged, which is bound
        # to onProviderStatusChanged.
        self.provider.prime(forwardHint)

    def showEvent(self, event: QShowEvent):
        self.provider.freeze(False)
        self.reevaluateSearchTerm()
        super().showEvent(event)

    def hideEvent(self, event: QHideEvent):
        self.provider.freeze(True)
        self.debounceTimer.stop()
        super().hideEvent(event)

    def hideOrBeep(self):
        if self.isVisible():  # close search bar if it doesn't have focus
            self.bail()
        else:
            QApplication.beep()

    def popUp(self, op: SearchBar.Op = Op.Start):
        wasHidden = self.isHidden()
        self.show()

        for button in self.buttons:
            button.setMaximumHeight(self.lineEdit.height())

        self.lineEdit.setFocus(Qt.FocusReason.PopupFocusReason)

        if op == SearchBar.Op.Start or wasHidden:
            self.lineEdit.selectAll()

        # Explicit jump next/previous
        if op == SearchBar.Op.Next:
            self.onEnterShortcut()
        elif op == SearchBar.Op.Previous:
            self.onShiftEnterShortcut()

    def bail(self):
        self.debounceTimer.stop()
        self.buddy.setFocus(Qt.FocusReason.PopupFocusReason)
        self.hide()

    def onSearchTextChanged(self, text: str):
        assert self.isVisible()
        self.hideToolTip()

        # Set text in provider
        self.provider.setTerm(text)

        # Schedule a debounce
        if self.provider.term():
            self.autoJumpWhenResultsComeIn = True
            self.debounceTimer.start()

    def _setProvider(self, provider: SearchProvider):
        self.lineEdit.clear()

        providerIndex = self.providers.index(provider)
        assert providerIndex >= 0, "unregistered provider"
        oldProvider = self.provider

        if provider is oldProvider:
            return

        action = next(a for a in self._providerChooserMenu.actions()
                      if a.data() is provider)
        action.setChecked(True)

        if oldProvider is not None:
            # Freeze it when not in use
            with QSignalBlockerContext(oldProvider):
                oldProvider.freeze(True)

        logger.debug(f"Swapping provider to {provider.shortTitle()}")

        self.provider = provider
        self.ui.filterCheckBox.setVisible(provider.canFilter())

        self.ui.providerChooser.setText(provider.shortTitle())
        self.ui.lineEdit.setPlaceholderText(provider.longTitle())

        provider.freeze(False)

        # Update styling
        self.autoJumpWhenResultsComeIn = False
        self.onProviderStatusChanged(provider.status())

    def onFilterCheckBoxToggled(self):
        assert self.isVisible()
        self.provider.setFilterState(self.ui.filterCheckBox.isChecked())

    def reevaluateSearchTerm(self):
        # Clear bad stem, if any
        self.provider.invalidate()

        if not self.isVisible():
            return

        # Re-evaluate search term. This will schedule a debounce,
        # but prevent the subsequent jump.
        self.onSearchTextChanged(self.rawSearchTerm)
        self.autoJumpWhenResultsComeIn = False

    def isRed(self) -> bool:
        return "true" == self.property("red")

    def onProviderStatusChanged(self, status: SearchProvider.TermStatus):
        self.hideToolTip()

        red = status == SearchProvider.TermStatus.Bad
        toggleQssProperty(self, "red", red)

        loupeIcon = "magnifying-glass-wait" if status == SearchProvider.TermStatus.Loading else "magnifying-glass"
        self.loupe.setIcon(stockIcon(loupeIcon))

        # If good search results just came in, jump to the next match
        # ...unless current selection already matches
        # ...unless an old search term is being reevaluated "passively"
        if (status == SearchProvider.TermStatus.Good
                and not self.provider.isCurrentMatch()
                and self.autoJumpWhenResultsComeIn):
            self.provider.jump(True)

    def showToolTip(self, text: str):
        # On macOS, Qt kills tooltips on keyup.
        # So, if enter key is down, stash tooltip until keyup.
        if SearchBar.MacToolTipQuirks:
            if self.trapNextToolTip:
                self.trappedToolTip = text
                return
            self.trappedToolTip = ""

        pos = QPoint(0, 0)

        # Hack: when the cursor is "above" the tooltip, then Qt makes the
        # tooltip vanish quickly. So try to avoid spawning the tooltip on top
        # of the cursor.
        if self.geometry().adjusted(0, 0, 0, 32).contains(self.mapFromGlobal(QCursor.pos())):
            pos.setY(-int(self.height() * 2.2))

        pos = self.mapToGlobal(pos)
        QToolTip.showText(pos, f"<p style='white-space: pre;'>{text}{self._ToolTipTag}", self)

    def hideToolTip(self):
        if SearchBar.MacToolTipQuirks:
            self.trappedToolTip = ""

        if QToolTip.isVisible() and QToolTip.text().endswith(self._ToolTipTag):
            QToolTip.hideText()

    if MacToolTipQuirks:
        # Schedule a tooltip that was trapped while the enter key was down.
        def keyReleaseEvent(self, event: QKeyEvent):
            super().keyReleaseEvent(event)
            self.trapNextToolTip = False
            if self.trappedToolTip:
                QTimer.singleShot(0, lambda t=self.trappedToolTip: self.showToolTip(t))

    @staticmethod
    def highlightNeedle(painter: QPainter, rect: QRect, text: str,
                        needlePos: int = 0, needleLen: int = -1,
                        widthUpToNeedle: int = -1, widthPastNeedle: int = -1,
                        lBleed: int = 2, rBleed: int = 2):
        """
        Helper function to render "found needle" text in the buddy widget's item
        delegate.
        """

        if needleLen < 0:
            needleLen = len(text)

        if widthUpToNeedle < 0 or widthPastNeedle < 0:
            fontMetrics = painter.fontMetrics()
            widthUpToNeedle = fontMetrics.horizontalAdvance(text, needlePos)
            widthPastNeedle = fontMetrics.horizontalAdvance(text, needlePos + needleLen)
        needleWidth = widthPastNeedle - widthUpToNeedle

        needleRect = QRect(rect.left() + widthUpToNeedle, rect.top(), needleWidth, rect.height())
        hiliteRect = QRectF(needleRect).marginsAdded(QMarginsF(lBleed, -1, rBleed, -1))

        painter.save()
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Yellow rounded rect
        path = QPainterPath()
        path.addRoundedRect(hiliteRect, 4, 4)
        painter.fillPath(path, colors.yellow)

        # Re-draw needle on top of highlight rect
        painter.setPen(Qt.GlobalColor.black)  # force black-on-yellow regardless of dark/light theme
        needleText = text[needlePos:needlePos + needleLen]
        painter.drawText(needleRect, Qt.AlignmentFlag.AlignCenter, needleText)

        painter.restore()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
from collections.abc import Iterable
from contextlib import suppress
from typing import Any

from gitfourchette import trtables
from gitfourchette import settings
from gitfourchette.gitdriver import GitDelta, GitDeltaSource, GitStatus
from gitfourchette.gitdriver.lfspointer import LfsPointerState
from gitfourchette.localization import *
from gitfourchette.nav import NavContext, NavLocator
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


def deltaModeText(om: FileMode, nm: FileMode) -> str:
    if nm == FileMode.UNREADABLE or nm == om:  # Deleted or unchanged mode
        return ""

    if nm == FileMode.BLOB and om == FileMode.BLOB_EXECUTABLE:
        return "-x"

    if nm == FileMode.BLOB_EXECUTABLE:
        return "+x"

    if nm == FileMode.LINK:
        return _("link")

    if nm == FileMode.TREE:
        return _("new subtree")

    if nm == FileMode.COMMIT:
        return _("commit in subtree")

    return ""


def fileTooltip(
        repo: Repo,
        delta: GitDelta,
        isCounterpart: bool = False
) -> str:
    locale = QLocale()
    of = delta.old
    nf = delta.new
    sc = delta.status

    text = "<table style='white-space: pre'>"

    def newLine(heading, caption):
        colon = _(':')
        color = mutedToolTipColorHex()
        return f"<tr><td style='color:{color}; text-align: right;'>{heading}{colon} </td><td>{caption}</td>"

    if sc == GitStatus.Renamed:
        text += newLine(_("old name"), escape(of.path))
        text += newLine(_("new name"), escape(nf.path))
    else:
        text += newLine(_("name"), escape(nf.path))

    # Status caption
    statusCaption = trtables.enum(sc)
    # Show status char except for untracked and conflict
    if sc not in [GitStatus.Untracked, GitStatus.Unmerged]:
        statusCaption += f" ({sc})"
    if sc == GitStatus.Unmerged:  # conflict sides
        assert delta.conflict is not None
        postfix = trtables.enum(delta.conflict.sides)
        statusCaption += f" ({postfix})"
    text += newLine(_("status"), statusCaption)

    # Similarity + Old name
    if sc == GitStatus.Renamed:
        text += newLine(_("similarity"), f"{delta.similarity}%")

    # File Mode
    if sc in [GitStatus.Deleted, GitStatus.Unmerged]:
        pass
    elif sc.isAddedOrUntracked:
        text += newLine(_("file mode"), trtables.enum(nf.mode))
    elif of.mode != nf.mode:
        text += newLine(_("file mode"), f"{trtables.enum(of.mode)} \u2192 {trtables.enum(nf.mode)}")

    # Get mtime & size
    mTimeNS, size = -1, -1
    sizeIsAccurate = False
    if nf.isBlob() and not nf.isId0():
        # Stat workdir file (get size on disk, modification time)
        if delta.source.isWorkdir():
            mTimeNS, size = nf.stat(repo)

        # Get accurate size in index if it's not an unstaged file
        if delta.source != GitDeltaSource.Dirty:
            assert nf.isIdValid()
            size = repo.peel_blob(nf.id).size
            sizeIsAccurate = True

    # Cache LFS pointer info
    if settings.prefs.lfsAware:
        delta.cacheLfsPointers(repo)

    # Size (if applicable)
    if size != -1:
        sizeText = locale.formattedDataSize(size, 1)
        if not sizeIsAccurate:
            sizeText = _("{size} on disk", size=sizeText)
        if delta.new.lfs.size >= 0 and not delta.new.lfs.isTentative():
            sizeText = f"{locale.formattedDataSize(delta.new.lfs.size, 1)} <b>(LFS)<b>"
        text += newLine(_("size"), sizeText)

    # Modified time
    if mTimeNS != -1:
        timeSecs = int(mTimeNS * 1e-9)  # Convert from nanoseconds
        timeQdt = QDateTime.fromSecsSinceEpoch(timeSecs)
        timeText = locale.toString(timeQdt, settings.prefs.shortTimeFormat)
        text += newLine(_("modified"), timeText)

    # Blob/Commit IDs
    if sc == GitStatus.Unmerged or nf.mode == FileMode.TREE:
        # Hide hashes for:
        # - unmerged conflicts
        # - untracked trees: those never have a valid ID
        pass
    elif of.lfs or nf.lfs:
        oldId = shortHash(of.lfs.id) if of.lfs.id else _("(not computed)")
        newId = shortHash(nf.lfs.id) if nf.lfs.id else _("(not computed)")
        text += newLine(_("LFS object hash"), f"{oldId} \u2192 {newId}")
    else:
        oldId = shortHash(of.id) if of.isIdValid() else _("(not computed)")
        newId = shortHash(nf.id) if nf.isIdValid() else _("(not computed)")
        idLegend = _("commit hash") if nf.mode == FileMode.COMMIT else _("blob hash")
        text += newLine(idLegend, f"{oldId} \u2192 {newId}")

    if isCounterpart:
        if delta.source == GitDeltaSource.Dirty:
            counterpartText = _("Currently viewing diff of staged changes in this file; "
                                "it also has <u>unstaged</u> changes.")
        else:
            counterpartText = _("Currently viewing diff of unstaged changes in this file; "
                                "it also has <u>staged</u> changes.")
        text += f"<p>{counterpartText}</p>"

    return text


class FileListModel(QAbstractListModel):
    class Role:
        Delta = Qt.ItemDataRole(Qt.ItemDataRole.UserRole + 0)
        FilePath = Qt.ItemDataRole(Qt.ItemDataRole.UserRole + 1)
        Locator = Qt.ItemDataRole(Qt.ItemDataRole.UserRole + 2)
        Decoration2 = Qt.ItemDataRole(Qt.ItemDataRole.UserRole + 3)  # LFS icons

    deltas: list[GitDelta]
    fileRows: dict[str, int]
    highlightedCounterpartRow: int

    repo: Repo

    navContext: NavContext
    """
    COMMITTED, STAGED or UNSTAGED.
    Does not change throughout the lifespan of this FileListModel.
    """

    navLocator: NavLocator
    """
    NavLocator for the commit that is currently being shown.
    Only valid (non-empty) if navContext == COMMITTED.
    Does not contain paths.
    """

    def __init__(self, parent: QWidget, repo: Repo, navContext: NavContext):
        super().__init__(parent)
        self.repo = repo
        self.navContext = navContext
        self.navLocator = NavLocator.Empty
        self.clear()

    @property
    def parentWidget(self) -> QWidget:
        parentWidget = self.parent()
        assert isinstance(parentWidget, QWidget)
        return parentWidget

    def clear(self):
        self.deltas = []
        self.fileRows = {}
        self.highlightedCounterpartRow = -1
        self.navLocator = NavLocator.Empty
        self.modelReset.emit()

    def setContents(self, deltas: Iterable[GitDelta]):
        self.beginResetModel()

        self.deltas.clear()
        self.fileRows.clear()

        sortedDeltas = sorted(deltas, key=lambda d: naturalSort(d.new.path))

        for delta in sortedDeltas:
            self.fileRows[delta.new.path] = len(self.deltas)
            self.deltas.append(delta)

        self.endResetModel()

    def rowCount(self, parent: QModelIndex = QModelIndex_default) -> int:
        return len(self.deltas)

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        row = index.row()
        try:
            delta = self.deltas[row]
        except IndexError:
            delta = None

        if role == FileListModel.Role.Delta:
            return delta

        elif role == FileListModel.Role.FilePath:
            # TODO: Canonical path for submodules?
            return delta.new.path

        elif role == FileListModel.Role.Locator:
            if self.navLocator:
                return self.navLocator.replace(path=delta.new.path)
            else:
                context = NavContext.fromGitDeltaSource(delta.new.source)
                return NavLocator(context, path=delta.new.path)

        elif role == Qt.ItemDataRole.DisplayRole:
            # TODO: Canonical path for submodules?
            text = abbreviatePath(delta.new.path, settings.prefs.pathDisplayStyle, allowNul=True)

            # Show important mode info in brackets
            modeInfo = deltaModeText(delta.old.mode, delta.new.mode)
            if modeInfo:
                text = f"[{modeInfo}] {text}"

            return text

        elif role == Qt.ItemDataRole.DecorationRole:
            return stockIcon(f"status_{delta.status.lower()}")

        elif role == FileListModel.Role.Decoration2:
            if settings.prefs.lfsAware:
                # Cache LFS pointer on demand as the file scrolls into view
                if delta.new.lfs.state == LfsPointerState.Unknown:
                    delta.cacheLfsPointers(self.repo)

                if delta.old.lfs and not delta.new.lfs:
                    return stockIcon("git-lfs-remove")
                elif delta.new.lfs.state != LfsPointerState.Valid:
                    pass
                elif delta.old.lfs and delta.new.lfs:
                    return stockIcon("git-lfs")
                elif not delta.old.lfs and delta.new.lfs:
                    return stockIcon("git-lfs-add")

        elif role == Qt.ItemDataRole.ToolTipRole:
            isCounterpart = row == self.highlightedCounterpartRow
            return fileTooltip(self.repo, delta, isCounterpart)

        elif role == Qt.ItemDataRole.FontRole:  # noqa: SIM102
            if row == self.highlightedCounterpartRow:
                font = self.parentWidget.font()
                font.setUnderline(True)
                return font

        return None

    def getRowForFile(self, path: str) -> int:
        """
        Get the row number for the given path.
        Raise KeyError if the path is absent from this model.
        """
        return self.fileRows[path]

    def getFileAtRow(self, row: int) -> str:
        """
        Get the path corresponding to the given row number.
        Return an empty string if the row number is invalid.
        """
        if row < 0 or row >= self.rowCount():
            return ""
        return self.data(self.index(row), FileListModel.Role.FilePath)

    def hasFile(self, path: str) -> bool:
        """
        Return True if the given path is present in this model.
        """
        return path in self.fileRows

    def matchPathspec(self, pattern: str) -> str:
        # Try new side first
        with suppress(StopIteration):
            return next(d.new.path for d in self.deltas if d.new.matchPathspec(pattern))

        # Try old side (but return path from *new* side)
        with suppress(StopIteration):  # type: ignore[unreachable]
            return next(d.new.path for d in self.deltas if d.old.matchPathspec(pattern))

        return ""



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

from typing import ClassVar

from gitfourchette.application import GFApplication
from gitfourchette.blameview.blamemodel import BlameModel, Revision
from gitfourchette.blameview.blamescrubber import BlameScrubber
from gitfourchette.blameview.blametextedit import BlameTextEdit
from gitfourchette.blameview.blamebusyspinner import BlameBusySpinner
from gitfourchette.graphview.commitlogmodel import CommitLogModel
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavHistory, NavFlags
from gitfourchette.qt import *
from gitfourchette.tasks import RepoTaskRunner
from gitfourchette.toolbox import *


class BlameWindow(QWidget):
    exploreCommit = Signal(NavLocator)

    _currentBlameWindows: ClassVar[list[BlameWindow]] = []
    "Currently open BlameWindows"

    def __init__(self, blameModel: BlameModel):
        # DON'T parent the window so it doesn't sit on top of MainWindow on Wayland
        super().__init__(None)

        # Keep reference around to avoid being GC'd instantly
        BlameWindow._currentBlameWindows.append(self)

        self.setObjectName("BlameWindow")

        self.model = blameModel
        self.taskRunner = RepoTaskRunner(self)
        self.navHistory = NavHistory()

        self.textEdit = BlameTextEdit(blameModel, self)
        self.textEdit.setUpAsDetachedWindow()

        self.scrubber = BlameScrubber(blameModel, self)
        self.scrubber.activated.connect(self.onScrubberActivated)

        self.jumpButton = QToolButton()
        self.jumpButton.setText(_("Jump"))
        self.jumpButton.setToolTip(_("View this commit in the repo"))
        self.jumpButton.setIcon(stockIcon("go-window@20px"))
        self.jumpButton.clicked.connect(lambda: self.jumpToCommit())

        oldNewTipTemplate = (f"<p style='white-space: pre'>{{0}}<br>"
                             f"<span style='color: {mutedToolTipColorHex()}'>({_('Shift+Click:')} {{1}})</span>")

        self.olderButton = QToolButton()
        self.olderButton.setText(_("Older"))
        self.olderButton.clicked.connect(self.goOlder)
        self.olderButton.setToolTip(oldNewTipTemplate.format(_("Go to next older revision"), _("Jump to bottom")))
        self.olderButton.setIcon(stockIcon("go-older"))

        self.newerButton = QToolButton()
        self.newerButton.setText(_("Newer"))
        self.newerButton.clicked.connect(self.goNewer)
        self.newerButton.setToolTip(oldNewTipTemplate.format(_("Go to next newer revision"), _("Jump to top")))
        self.newerButton.setIcon(stockIcon("go-newer"))

        self.backButton = QToolButton()
        self.backButton.setText(_("Back"))
        self.backButton.clicked.connect(self.goBack)
        self.backButton.setToolTip(_("Navigate back"))
        self.backButton.setIcon(stockIcon("back"))

        self.forwardButton = QToolButton()
        self.forwardButton.setText(_("Forward"))
        self.forwardButton.clicked.connect(self.goForward)
        self.forwardButton.setToolTip(_("Navigate forward"))
        self.forwardButton.setIcon(stockIcon("forward"))

        topBar = QToolBar()
        topBar.setIconSize(QSize(20, 20))
        topBar.addWidget(self.backButton)
        topBar.addWidget(self.forwardButton)
        topBar.addSeparator()
        topBar.addWidget(self.newerButton)
        topBar.addWidget(self.olderButton)
        topBar.addWidget(self.scrubber)
        topBar.addWidget(self.jumpButton)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(QMargins())
        layout.addWidget(topBar)
        layout.addWidget(self.textEdit.searchBar)
        layout.addWidget(self.textEdit)
        layout.setSpacing(0)

        self.textEdit.showRevision.connect(self.showRevision)
        self.textEdit.jumpToCommit.connect(self.jumpToCommit)

        self.setWindowModality(Qt.WindowModality.NonModal)
        self.setWindowFlags(Qt.WindowType.Window)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

        self.busySpinner = BlameBusySpinner(self.textEdit)
        self.textEdit.sizeChanged.connect(self.busySpinner.reposition)

        self.resize(self.textEdit.idealDetachedSize())

        app = GFApplication.instance()
        app.mouseSideButtonPressed.connect(self.onMouseSideButtonPressed)

    # -------------------------------------------------------------------------
    # Event filters & handlers

    def onMouseSideButtonPressed(self, forward: bool):
        if not self.isActiveWindow():
            return
        delta = [-1, 1][forward]
        self.goBackOrForwardDelta(delta)

    def closeEvent(self, event: QCloseEvent):
        assert self in self._currentBlameWindows
        self._currentBlameWindows.remove(self)

        self.scrubber.scrubberDelegate.prepareForDeletion()
        self.scrubber.deleteLater()

        self.taskRunner.prepareForDeletion()
        self.model.prepareForDeletion()

        super().closeEvent(event)

    # -------------------------------------------------------------------------

    def saveFilePosition(self) -> NavLocator:
        currentLocator = self.textEdit.preciseLocator()
        self.navHistory.push(currentLocator)
        return currentLocator

    def showRevision(self, revision: Revision, saveAndTransposePosition=True):
        from gitfourchette.tasks.blametasks import BlameRevision
        BlameRevision.invoke(self, revision, saveAndTransposePosition)

    def syncNavButtons(self):
        index = self.scrubber.currentIndex()
        count = self.scrubber.count()

        self.newerButton.setEnabled(index != 0 and count >= 2)
        self.olderButton.setEnabled(index != count - 1)

        self.backButton.setEnabled(self.navHistory.canGoBack())
        self.forwardButton.setEnabled(self.navHistory.canGoForward())

    def goNewer(self):
        if QGuiApplication.keyboardModifiers() == Qt.KeyboardModifier.ShiftModifier:
            topRevision = self.model.revList.sequence[0]
            self.showRevision(topRevision)
        else:
            self.goNewerOrOlder(-1)

    def goOlder(self):
        if QGuiApplication.keyboardModifiers() == Qt.KeyboardModifier.ShiftModifier:
            bottomRevision = self.model.revList.sequence[-1]
            self.showRevision(bottomRevision)
        else:
            self.goNewerOrOlder(1)

    def goBack(self):
        self.goBackOrForwardDelta(-1)

    def goForward(self):
        self.goBackOrForwardDelta(1)

    def goNewerOrOlder(self, delta: int):
        index = self.scrubber.currentIndex()
        index += delta
        assert 0 <= index < self.scrubber.count()
        revision = self.getRevisionFromScrubberRow(index)
        self.showRevision(revision)

    def goBackOrForwardDelta(self, delta: int):
        self.saveFilePosition()
        locator = self.navHistory.navigateDelta(delta)
        if not locator:
            return
        revision = self.model.revList.revisionForCommit(locator.commit)
        self.showRevision(revision, saveAndTransposePosition=False)

    def onScrubberActivated(self, index: int):
        revision = self.getRevisionFromScrubberRow(index)
        self.showRevision(revision)

    def getRevisionFromScrubberRow(self, index: int) -> Revision:
        return self.scrubber.itemData(index, CommitLogModel.Role.BlameRevision)

    def jumpToCommit(self, locator: NavLocator = NavLocator.Empty):
        if locator == NavLocator.Empty:
            locator = self.model.currentRevision.toLocator()
        locator = locator.withExtraFlags(NavFlags.ActivateWindow)
        self.exploreCommit.emit(locator)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import logging
import os
import shlex
from contextlib import suppress
from pathlib import Path
from signal import SIGINT

from gitfourchette import trtables
from gitfourchette.application import GFApplication
from gitfourchette.exttools.toolcommands import ToolCommands
from gitfourchette.exttools.toolpresets import ToolPresets
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.settings import prefs
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


def onLocateTool(prefKey: str, newPath: str):
    command = getattr(prefs, prefKey)
    newCommand = ToolCommands.replaceProgramTokenInCommand(command, newPath)
    setattr(prefs, prefKey, newCommand)
    prefs.write()


def onExternalToolProcessError(parent: QWidget, prefKey: str,
                               isKnownFlatpak=False,
                               compileError: Exception | None = None):
    assert isinstance(parent, QWidget)

    commandString = getattr(prefs, prefKey)
    programName = ToolPresets.getCommandName(commandString)

    translatedPrefKey = trtables.prefKey(prefKey)

    title = _("Failed to start {tool}", tool=translatedPrefKey)

    if compileError:
        message = _("There is an issue with the {tool} command template in your settings:").format(tool=tquo(translatedPrefKey))
        message += f"<pre>{escape(commandString)}</pre>"
        message += "&rarr; " + str(compileError)
    else:
        if isKnownFlatpak:
            message = _("Couldn’t start Flatpak {command} ({tool}).")
        else:
            message = _("Couldn’t start {command} ({tool}).")
        message = message.format(tool=translatedPrefKey, command=bquo(programName))
        message += " " + _("It might not be installed on your machine.")

    configureButtonID = QMessageBox.StandardButton.Ok
    browseButtonID = QMessageBox.StandardButton.Open

    qmb = asyncMessageBox(parent, 'warning', title, message,
                          configureButtonID | browseButtonID | QMessageBox.StandardButton.Cancel,
                          deleteOnClose=True)

    def browseDialog():
        qfd = QFileDialog(parent, _("Where is {tool}?", tool=lquo(programName)))
        qfd.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        qfd.setFileMode(QFileDialog.FileMode.AnyFile)
        qfd.setWindowModality(Qt.WindowModality.WindowModal)
        qfd.setOption(QFileDialog.Option.DontUseNativeDialog, APP_TESTMODE)
        qfd.show()
        qfd.fileSelected.connect(lambda newPath: onLocateTool(prefKey, newPath))
        return qfd

    configureButton = qmb.button(configureButtonID)
    configureButton.setText(_("Edit Command…"))
    configureButton.setIcon(stockIcon("configure"))
    configureButton.clicked.connect(lambda: GFApplication.instance().openPrefsDialog(prefKey))

    browseButton = qmb.button(browseButtonID)
    browseButton.setText(_("Locate {tool}…", tool=lquo(programName)))
    browseButton.clicked.connect(browseDialog)

    removeBrowseButton = bool(compileError)
    if FREEDESKTOP:
        tokens = ToolCommands.splitCommandTokens(commandString)
        if ToolCommands.isFlatpakRunCommand(tokens):
            removeBrowseButton = True
    if removeBrowseButton:
        qmb.removeButton(browseButton)

    qmb.show()


def setUpToolCommand(parent: QWidget, prefKey: str):
    translatedPrefKey = trtables.prefKey(prefKey)

    title = translatedPrefKey

    message = _("{tool} isn’t configured in your settings yet.", tool=bquo(translatedPrefKey))

    qmb = asyncMessageBox(parent, 'warning', title, message,
                          QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)

    configureButton = qmb.button(QMessageBox.StandardButton.Ok)
    configureButton.setText(_("Set up {tool}", tool=lquo(translatedPrefKey)))

    qmb.accepted.connect(lambda: GFApplication.instance().openPrefsDialog(prefKey))
    qmb.show()


class ToolProcess(QProcess):
    PrefKeyEditor = "externalEditor"
    PrefKeyDiffTool = "externalDiff"
    PrefKeyMergeTool = "externalMerge"

    def __init__(self, parent: QWidget, tokens: list[str], directory: str, detached: bool, prefKey: str, environment: dict[str, str] | None = None):
        super().__init__(parent)

        self.detached = detached
        self.prefKey = prefKey

        self.setProgram(tokens[0])
        self.setArguments(tokens[1:])
        self.setWorkingDirectory(directory)
        self.setProcessChannelMode(QProcess.ProcessChannelMode.ForwardedChannels)
        ToolCommands.setQProcessEnvironment(self, environment)

        ToolCommands.wrapFlatpakCommand(self, detached=detached)

        # Listen for process completion.
        self.finished.connect(self.onFinished)
        # Listen for errors (typically FailedToStart).
        self.errorOccurred.connect(self.onErrorOccurred)
        # When the parent dies, don't let callbacks hit zombie objects.
        parent.destroyed.connect(self.onParentDestroyedWhileProcessRunning)

        # Decide whether to actually start detached
        if FLATPAK:
            # Flatpaks detach from process via ToolCommands.compileCommand
            startDetached = False
        elif MACOS:
            startDetached = detached and tokens[0] != "/usr/bin/open"
        else:
            startDetached = detached

        logger.info("Process starting: " + shlex.join([self.program()] + self.arguments()))
        if startDetached:
            self.startDetached()
        else:
            self.start()
            logger.info(f"(PID {self.processId()})")

    def parentWidget(self) -> QWidget:
        parent = self.parent()
        assert isinstance(parent, QWidget)
        return parent

    def onFinished(self, code: int, status: QProcess.ExitStatus):
        logger.info(f"Process done: {code} {status}")
        self.disconnectFromParent()
        if not WINDOWS and code == 127:
            # The Flatpak distribution runs non-Flatpak commands through `env`,
            # which returns 127 if the command isn't found.
            onExternalToolProcessError(self.parentWidget(), self.prefKey)

    def onErrorOccurred(self, processError: QProcess.ProcessError):
        logger.info(f"Process error: {processError}")
        self.disconnectFromParent()
        onExternalToolProcessError(self.parentWidget(), self.prefKey)

    def onParentDestroyedWhileProcessRunning(self):
        # Disconnect signals
        self.disconnectFromParent()
        self.finished.disconnect(self.onFinished)
        self.errorOccurred.disconnect(self.onErrorOccurred)

        if not self.detached:
            pid = self.processId()
            if pid > 0 and not WINDOWS:
                # SIGINT works better with non-Flatpak Meld
                os.kill(pid, SIGINT)
            else:
                ToolCommands.terminatePlus(self)

        self.setParent(None)
        self.deleteLater()

    def disconnectFromParent(self):
        # Disconnect from the parent so that future signals won't touch a parent
        # that might have been destroyed.
        # Ignore TypeError, raised by PyQt6 if the connection has already been
        # undone (may occur when SIGKILLing the process - both onFinished and
        # onErrorOccurred will fire).
        with suppress(TypeError):
            self.parent().destroyed.disconnect(self.onParentDestroyedWhileProcessRunning)

    @classmethod
    def startProcess(
            cls,
            parent: QWidget,
            prefKey: str,
            replacements: dict[str, str],
            positional: list[str],
            allowQDesktopFallback: bool = False,
            directory: str = "",
            detached: bool = False,
            environment: dict[str, str] | None = None
    ) -> ToolProcess | None:
        assert isinstance(parent, QWidget)

        command = getattr(prefs, prefKey, "").strip()

        if not command and allowQDesktopFallback:
            for argument in positional:
                QDesktopServices.openUrl(QUrl.fromLocalFile(argument))
            return None

        if not command:
            setUpToolCommand(parent, prefKey)
            return None

        try:
            tokens, directory = ToolCommands.compileCommand(command, replacements, positional, directory, detached)
        except ValueError as exc:
            onExternalToolProcessError(parent, prefKey, compileError=exc)
            return None

        # Check if the Flatpak is installed
        if FREEDESKTOP:
            # Check 'isFlatpakRunCommand' on unfiltered tokens (compileCommand may have added a wrapper)
            unfilteredTokens = ToolCommands.splitCommandTokens(command)
            flatpakRefTokenIndex = ToolCommands.isFlatpakRunCommand(unfilteredTokens)
            if flatpakRefTokenIndex and not ToolCommands.isFlatpakInstalled(unfilteredTokens[flatpakRefTokenIndex]):
                onExternalToolProcessError(parent, prefKey, isKnownFlatpak=True)
                return None

        process = ToolProcess(parent=parent, tokens=tokens, directory=directory, detached=detached, prefKey=prefKey, environment=environment)
        return process

    @classmethod
    def startTextEditor(cls, parent: QWidget, path: str):
        return cls.startProcess(
            parent,
            cls.PrefKeyEditor,
            positional=[path],
            replacements={},
            allowQDesktopFallback=True)

    @classmethod
    def startDiffTool(cls, parent: QWidget, a: str, b: str):
        return cls.startProcess(
            parent,
            cls.PrefKeyDiffTool,
            positional=[],
            replacements={"$L": a, "$R": b})

    @classmethod
    def startTerminal(cls, parent: QWidget, workdir: str, commandTokens: list[str] | None = None):
        if commandTokens is None:
            commandTokens = []

        launcherScriptPath = ToolCommands.makeTerminalScript(workdir, commandTokens)

        if WINDOWS:  # Compatibility with bash.exe
            launcherScriptPath = Path(launcherScriptPath).as_posix()

        return cls.startProcess(
            parent=parent,
            prefKey="terminal",
            replacements={"$COMMAND": launcherScriptPath},
            positional=[],
            directory=workdir,
            detached=True)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Tasks that navigate to a specific area of the repository.

Unlike most other tasks, jump tasks directly manipulate the UI extensively, via RepoWidget.
"""

import dataclasses
import logging
import os
import re

from gitfourchette import trtables
from gitfourchette import settings
from gitfourchette.diffview.diffdocument import DiffDocument
from gitfourchette.diffview.specialdiff import SpecialDiffError, ImageDelta
from gitfourchette.gitdriver import GitConflict, GitDelta, GitStatus, GitDriver, argsIf
from gitfourchette.gitdriver.parsers import parseAheadBehind
from gitfourchette.graphview.commitlogmodel import SpecialRow
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavContext, NavFlags
from gitfourchette.porcelain import NULL_OID, Oid, commit_diff_pair
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEREF, UC_FAKEID
from gitfourchette.tasks import TaskPrereqs
from gitfourchette.tasks.loadtasks import LoadPatch, TAbstractDiffDocument
from gitfourchette.tasks.repotask import AbortTask, RepoTask, TaskEffects, RepoGoneError
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)

_submoduleIndexLinePattern = re.compile(r"^index ([\da-f]+)\.\.([\da-f]+)", re.MULTILINE)


def loadWorkdir(task: RepoTask, allowWriteIndex: bool):
    """
    Refresh staged/dirty GitDeltas in the RepoModel.
    """

    # Get oid of the head commit (or None if unborn).
    # It will be stored in GitDeltaFile.old in staged files.
    if task.repo.head_is_unborn:
        headCommitId = None
    else:
        headCommitId = task.repo.head_commit_id

    # Run 'git status'
    gitStatus = yield from task.flowCallGit(
        *argsIf(not allowWriteIndex, "--no-optional-locks"),
        "status",
        "--porcelain=v2",
        "-z",
        "--untracked-files=all")

    # Get GitDelta lists from 'git status' output
    numEntries, stagedDeltas, unstagedDeltas = gitStatus.readStatusPorcelainV2Z(headCommitId)

    if APP_DEBUG:
        assert not any(d.submoduleStatus.startswith("S") for d in stagedDeltas), "only expecting full submo status in unstaged deltas"

    # Pre-cache LFS state for unstaged files via `git check-attr` if
    # .gitattributes has any unstaged changes.
    # LFS state is usually queried via libgit2's git_get_attr. However, it's
    # unreliable when .gitattributes has unstaged changes, because it falls back
    # to the indexed revision of .gitattributes when an attr was deleted from
    # the file but still exists in the index. Ideally, libgit2 should offer a
    # GIT_ATTR_CHECK_FILE_ONLY flag so we don't have to do this.
    with Benchmark("check-attr unstaged"):
        if any(d.new.path.endswith(".gitattributes") for d in unstagedDeltas):
            unstagedTable = {d.new.path: d for d in unstagedDeltas}

            checkAttrDriver = yield from task.flowCallGit(
                "check-attr", "-z", "filter", "--", *unstagedTable.keys())

            # Can't use stdoutTable because all fields are zero-separated.
            tokens = checkAttrDriver.stdoutScrollback().removesuffix("\0").split("\0")
            for i in range(0, len(tokens), 3):
                # (filename, attribute, unspecified/unset/set/value) triplets
                fileName = tokens[i]
                value = tokens[i + 2]
                delta = unstagedTable[fileName]
                delta.new.cacheLfsPointer(task.repo, value)

    # Fill in submodule commit hashes.
    # (This might be parallelizable if we've got tons of modified submodules)
    for delta in unstagedDeltas:
        # Scan for submodules with changes that the superproject isn't tracking yet
        submoduleUpdated = delta.submoduleStatus.startswith("S")
        if not submoduleUpdated:
            continue

        if delta.status == GitStatus.Deleted:
            assert delta.new.isId0()
            continue

        # Work out head commit for this submodule
        headDidMove = "C" in delta.submoduleStatus
        if not headDidMove:
            # The submodule's head hasn't moved.
            submoduleCommitHash = delta.old.id
        else:
            # The submodule's head has moved, but we don't know to what commit,
            # because "git status" doesn't give this information if unstaged.
            subDiffDriver = yield from task.flowCallGit("diff", "--submodule=short", "--full-index", delta.new.path)
            subDiff = subDiffDriver.stdoutScrollback()
            match = _submoduleIndexLinePattern.search(subDiff)
            assert match
            assert delta.old.id == match.group(1)
            submoduleCommitHash = match.group(2)

        # Fill in worktree hash
        assert not delta.new.isIdValid(), f"not expecting id to be filled in right after git status: {delta.new.id}"
        delta.new.id = submoduleCommitHash

    repoModel = task.repoModel
    repoModel.workdirUnstagedDeltas = unstagedDeltas
    repoModel.workdirStagedDeltas = stagedDeltas
    repoModel.workdirNumChanges = numEntries
    repoModel.workdirStatusReady = True

    # Update pathspec filter
    cpf = repoModel.commitPathspecFilter
    if not cpf.needle:
        pass
    elif repoModel.workdirMatchesPathNeedle(cpf.needle):
        if UC_FAKEID not in cpf.matchingIds:
            cpf.matchingIds.add(UC_FAKEID)
            cpf.resultsUpdated.emit()
    else:
        if UC_FAKEID in cpf.matchingIds:
            cpf.matchingIds.remove(UC_FAKEID)
            cpf.resultsUpdated.emit()

    # Refresh libgit2 index after git status is complete
    repoModel.repo.refresh_index()


class Jump(RepoTask):
    """
    Single entry point to navigate to any NavLocator in a repository.

    This is the only task that is allowed to change the value of RepoWidget.navLocator.
    """

    @dataclasses.dataclass
    class Result(Exception):
        locator: NavLocator
        document: TAbstractDiffDocument | None
        delta: GitDelta | None = None

    def canKill(self, task: RepoTask):
        return isinstance(task, Jump
                          | JumpBack
                          | JumpForward
                          | JumpToHEAD
                          | JumpToUncommittedChanges
                          | RefreshRepo)

    def flow(self, locator: NavLocator):
        if not locator:
            return

        # Back up current locator
        self.rw.saveFilePositions()

        result = yield from self.loadResult(locator)

        assert onAppThread()
        self.saveFinalLocator(result.locator)
        self.displayResult(result)

        if locator.hasFlags(NavFlags.ActivateWindow):  # initial locator!
            self.rw.activateWindow()

    def loadResult(self, locator: NavLocator) -> RepoTask.Flow[Result]:
        rw = self.rw

        # If the locator is "coarse" (i.e. no specific path given, just a generic context),
        # try to recall where we were last time we looked at this context.
        locator = rw.navHistory.refine(locator)

        try:
            # Load workdir or commit, and show the corresponding view
            if locator.context == NavContext.SPECIAL:
                self.showSpecial(locator)  # always raises Jump.Result
            elif locator.context.isWorkdir():
                locator = yield from self.showWorkdir(locator)
            else:
                locator = yield from self.showCommit(locator)
        except Jump.Result as result:
            # The showXXX functions may bail early by raising Jump.Result.
            # Set up DiffArea for this locator.
            rw.diffArea.setUpForLocator(result.locator)
            return result

        fileList = rw.diffArea.fileListByContext(locator.context)

        # If we don't have a path in the locator, fall back to first path in file list.
        # (Only for non-special locators, though!)
        if not locator.path and locator.context != NavContext.SPECIAL:
            locator = locator.replace(path=fileList.firstPath())
            locator = rw.navHistory.refine(locator)

        # Set up DiffArea for this locator.
        # Note that this may return a new locator if the desired path is not available.
        locator = rw.diffArea.setUpForLocator(locator)

        # Blank path?
        if not locator.path:
            return Jump.Result(locator, None)

        delta = fileList.deltaForFile(locator.path)
        assert locator.context == NavContext.fromGitDeltaSource(delta.source)

        # If DiffView is already set up to display this specific patch,
        # we don't need to bother shelling out to 'git diff'.
        # TODO: Would also be nice for image diffs!
        if self.isDiffViewAlreadySetUpFor(locator, delta):
            return Jump.Result(rw.diffView.currentLocator, rw.diffView.currentDiffDocument, delta)

        # Load the patch
        diffDocument = yield from self.flowSubtask(LoadPatch, delta, locator)
        return Jump.Result(locator, diffDocument, delta)

    def isDiffViewAlreadySetUpFor(self, locator: NavLocator, delta: GitDelta) -> bool:
        currentLocator = self.rw.diffView.currentLocator
        currentDelta = self.rw.diffView.currentDelta

        # Special flag to bypass same-patch detection
        if locator.hasFlags(NavFlags.ForceRecreateDocument):
            return False

        # Coarse NavLocators must match
        if not locator.isSimilarEnoughTo(currentLocator):
            return False

        # Special case for unstaged files: a valid hash may not be available
        # in delta.new, so we rely on the file's stats on disk.
        if locator.context == NavContext.UNSTAGED:
            if delta.new.path != currentDelta.new.path:
                return False

            # Refresh filesystem status for unstaged files. Since `git status`
            # doesn't hash unstaged files, we rely on st_mtime_ns to determine if
            # we're displaying stale contents.
            if currentDelta.new.diskStat != currentDelta.new.stat(self.repo):
                return False

            # If all else is equal, equality hinges on the old delta.
            # Ignore the new delta, because DiffView may have a valid new hash.
            # We can expect the old hash (i.e. the blob in the index) to be valid.
            assert delta.old.isIdValid()
            return delta.old == currentDelta.old

        # In staged or commit contexts, we've got valid hashes for both sides.
        assert delta.old.isIdValid()
        assert delta.new.isIdValid()
        assert currentDelta.old.isIdValid()
        assert currentDelta.new.isIdValid()

        return delta == currentDelta

    def showWorkdir(self, locator: NavLocator) -> RepoTask.Flow[NavLocator]:
        rw = self.rw
        repoModel = self.repoModel

        # Save selected row number for the end of the function
        previousRowStaged = rw.stagedFiles.earliestSelectedRow()
        previousRowDirty = rw.dirtyFiles.earliestSelectedRow()

        with (
            QSignalBlockerContext(rw.sidebar),  # Don't emit jump signals
            QScrollBackupContext(rw.sidebar),  # Stabilize scroll bar value
        ):
            rw.sidebar.selectAnyRef(UC_FAKEREF)

        self.showLocatorInGraphView(locator)

        # Reset diff banner
        rw.diffArea.diffBanner.setVisible(False)
        rw.diffArea.contextHeader.setContext(locator)

        # Stale workdir model - force load workdir
        forceDiff = locator.hasFlags(NavFlags.ForceDiff)
        writeIndex = locator.hasFlags(NavFlags.AllowWriteIndex)
        if forceDiff or repoModel.workdirStale:
            # Load workdir (async)
            if forceDiff or not repoModel.workdirStatusReady:
                yield from loadWorkdir(self, allowWriteIndex=writeIndex)

            # Fill FileListViews
            with QSignalBlockerContext(rw.dirtyFiles, rw.stagedFiles):  # Don't emit jump signals
                rw.dirtyFiles.setContents(repoModel.workdirUnstagedDeltas)
                rw.stagedFiles.setContents(repoModel.workdirStagedDeltas)

            # Invalidate ConflictView if its current conflict is gone
            cc = rw.conflictView.currentConflict
            if not any(cc == d.conflict for d in repoModel.workdirUnstagedDeltas):
                rw.conflictView.invalidate()

            # Update file counts in captions
            nDirty = rw.dirtyFiles.model().rowCount()
            nStaged = rw.stagedFiles.model().rowCount()
            rw.diffArea.dirtyHeader.setText(_n("Unstaged ({n})", "Unstaged ({n})", nDirty))
            rw.diffArea.stagedHeader.setText(_n("Staged ({n})", "Staged ({n})", nStaged))
            rw.diffArea.commitButton.setText(_n("Commit {n} file", "Commit {n} files", nStaged))

            # Make Commit button bold if anything is staged
            commitButtonFont = rw.diffArea.commitButton.font()
            commitButtonBold = nStaged != 0
            if commitButtonFont.bold() != commitButtonBold:
                commitButtonFont.setBold(commitButtonBold)
                rw.diffArea.commitButton.setFont(commitButtonFont)

            # Flip workdir freshness
            repoModel.workdirStale = False
            repoModel.workdirStatusReady = True

        # If jumping to generic workdir context, find a concrete context
        if locator.context == NavContext.WORKDIR:
            if rw.dirtyFiles.isEmpty() and not rw.stagedFiles.isEmpty():
                locator = locator.replace(context=NavContext.STAGED)
            else:
                locator = locator.replace(context=NavContext.UNSTAGED)
            locator = rw.navHistory.refine(locator)

        # Early out if workdir is clean
        if rw.dirtyFiles.isEmpty() and rw.stagedFiles.isEmpty():
            locator = locator.replace(path="")
            sde = SpecialDiffError(
                _("The working directory is clean."),
                _("There aren’t any changes to commit."))
            raise Jump.Result(locator, sde)

        assert not locator.hasFlags(NavFlags.FuzzyPath), "FuzzyPath should not occur in the workdir"

        # (Un)Staging a file makes it vanish from its file list.
        # But we don't want the selection to go blank in this case.
        # Restore selected row (by row number) in the file list so the user
        # can keep hitting RETURN/DELETE to stage/unstage a series of files.
        isStaged = locator.context == NavContext.STAGED
        flModel = (rw.stagedFiles if isStaged else rw.dirtyFiles).flModel
        flPrevRow = previousRowStaged if isStaged else previousRowDirty

        if locator.path and not flModel.hasFile(locator.path) and flPrevRow >= 0:
            path = flModel.getFileAtRow(min(flPrevRow, flModel.rowCount()-1))
            locator = locator.replace(path=path)
            locator = locator.coarse(keepFlags=True)  # don't carry cursor over from old locator
            locator = rw.navHistory.refine(locator)

        return locator

    def showSpecial(self, locator: NavLocator):
        rw = self.rw
        locale = QLocale()

        with QSignalBlockerContext(rw.sidebar, rw.committedFiles):
            rw.sidebar.clearSelection()
            rw.diffArea.committedFiles.clear()
            rw.diffArea.committedHeader.setText(" ")
            rw.diffArea.diffBanner.hide()
            rw.diffArea.contextHeader.setContext(locator)

        self.showLocatorInGraphView(locator)

        special = SpecialRow.fromString(locator.path)

        if special == SpecialRow.EndOfShallowHistory:
            sde = SpecialDiffError(
                _("Shallow clone – End of available history."),
                _("More commits may be available in a full clone."))

        elif special == SpecialRow.TruncatedHistory:
            from gitfourchette import settings
            expandSome = makeInternalLink("expandlog")
            expandAll = makeInternalLink("expandlog", n=str(0))
            changePref = makeInternalLink("prefs", "maxCommits")
            humanNextThreshold = locale.toString(self.repoModel.nextTruncationThreshold)
            humanPrefThreshold = locale.toString(settings.prefs.maxCommits)
            options = [
                linkify(_("Load up to {0} commits", humanNextThreshold), expandSome),
                linkify(_("[Load full commit history] (this may take a moment)"), expandAll),
                linkify(_("[Change threshold setting] (currently: {0} commits)", humanPrefThreshold), changePref),
            ]
            sde = SpecialDiffError(
                _("History truncated to {0} commits.", locale.toString(self.repoModel.numRealCommits)),
                _("More commits may be available."),
                longform=toRoomyUL(options))

        elif special == SpecialRow.CannotCompareRows:
            sde = SpecialDiffError(
                _("The selected items cannot be compared."),
                icon="SP_MessageBoxWarning")

        elif special == SpecialRow.TooManyRowsSelected:
            controlKey = QKeySequence(Qt.Key.Key_Control).toString(QKeySequence.SequenceFormat.NativeText)
            numRows = len(locator.selectedCommits)
            sde = SpecialDiffError(
                _n("{n} item selected", "{n} items selected", n=numRows),
                _("You can compare up to two commits at a time. Tip: hold {0} "
                  "while making a selection to compare 2 discontiguous commits.", controlKey))

        else:
            raise NotImplementedError(f"Unsupported special locator: {special}")

        raise Jump.Result(locator, sde)

    def showCommit(self, locator: NavLocator) -> RepoTask.Flow[NavLocator]:
        """
        Jump to a commit.
        Return a refined NavLocator.
        """

        rw = self.rw
        area = rw.diffArea
        assert locator.context == NavContext.COMMITTED

        # If it's a ref, look it up
        if locator.ref:
            assert locator.commit == NULL_OID
            try:
                oid = self.repoModel.refs[locator.ref]
                locator = locator.replace(commit=oid, ref="")
            except KeyError as exc:
                raise AbortTask(_("Unknown reference {0}.", tquo(locator.ref))) from exc

        assert locator.commit
        assert not locator.ref

        try:
            stashIndex = rw.repoModel.stashes.index(locator.commit)
            isStash = True
        except ValueError:
            stashIndex = -1
            isStash = False

        commit = rw.repo.peel_commit(locator.commit)

        # Attempt to select matching ref in sidebar
        with (
            QSignalBlockerContext(rw.sidebar),  # Don't emit jump signals
            QScrollBackupContext(rw.sidebar),  # Stabilize scroll bar value
        ):
            if isStash:
                rw.sidebar.selectAnyRef(f"stash@{{{stashIndex}}}")
            else:
                refCandidates = rw.repoModel.refsAt.get(locator.commit, [])
                rw.sidebar.selectAnyRef(*refCandidates)

        flv = area.committedFiles
        area.diffBanner.setVisible(False)
        area.contextHeader.setContext(locator, commit.message, isStash)

        # Attempt to show the commit in GraphView
        self.showLocatorInGraphView(locator, isStash=isStash)

        if (not locator.hasFlags(NavFlags.ForceDiff)
                and locator.commit == rw.navLocator.commit
                and locator.selectedCommits == rw.navLocator.selectedCommits):
            # No need to reload the same commit diff
            logger.debug("Don't reload same commit diff")

        else:
            # Loading a different commit
            area.diffBanner.lastWarningWasDismissed = False

            # Load commit
            diffAB = locator.commitDiffAB()
            if not diffAB:
                diffAB = commit_diff_pair(commit)
            tokens = GitDriver.buildDiffRawCommand(diffAB)
            driver = yield from self.flowCallGit(*tokens)
            deltas = driver.readDiffRawZ()

            # Fill out source commits
            for d in deltas:
                d.old.sourceCommit = diffAB[0]
                d.new.sourceCommit = diffAB[1]

            summary = self.repo.peel_commit(locator.commit).message.strip()

            # Fill committed file list
            with QSignalBlockerContext(flv):  # Don't emit jump signals
                flv.clear()
                flv.setCommitLocator(locator)
                flv.setContents(deltas)
                numChanges = flv.model().rowCount()

            # Set header text
            headerText = toLengthVariants(_n("{n} change:|{n} ch.:", "{n} changes:|{n} ch.:", numChanges))
            area.committedHeader.setText(headerText)
            area.committedHeader.setToolTip("<p>" + escape(summary).replace("\n", "<br>"))

        # Early out if the commit is empty
        if flv.isEmpty():
            if locator.commitDiffAB():
                a, b = locator.commitDiffAB()
                message = _("No changes from {0} to {1}.", hquo(shortHash(a)), hquo(shortHash(b)))
                details = _("The trees are identical at both commits.")
            else:
                message = _("This commit is empty.")
                details = _("Commit {0} doesn’t affect any files.", hquo(shortHash(locator.commit)))
            sde = SpecialDiffError(message, details)
            locator = locator.replace(path="")
            raise Jump.Result(locator, sde)

        # Try to resolve a fuzzy path
        if locator.path and locator.hasFlags(NavFlags.FuzzyPath):
            resolvedPath = flv.flModel.matchPathspec(locator.path)
            locator = locator.replace(path=resolvedPath)
            locator = locator.withoutFlags(NavFlags.FuzzyPath)

        return locator

    def saveFinalLocator(self, locator: NavLocator):
        # Before saving the locator in the RepoWidget, strip one-time flags.
        locator = locator.withoutFlags(~NavFlags.KeepFlagsOnRefresh)

        self.rw.navLocator = locator

        self.rw.navHistory.push(locator)
        self.rw.historyChanged.emit()

    def showLocatorInGraphView(self, locator: NavLocator, isStash=False):
        graphView = self.rw.graphView

        with QSignalBlockerContext(graphView):  # Don't emit jump signals
            try:
                graphView.selectRowForLocator(locator)
            except graphView.SelectCommitError as e:
                # Commit is hidden or not loaded
                graphView.clearSelection()

                # Warning banner
                banner = self.rw.diffArea.diffBanner
                if not banner.lastWarningWasDismissed and not isStash:
                    warningText = str(e)
                    banner.popUp("", warningText, canDismiss=True, withIcon=True)

    def displayResult(self, result: Result):
        document = result.document
        area = self.rw.diffArea

        # Set header
        header = self.makeHeaderText(result.document, result.delta, result.locator)
        area.diffHeader.setText(header)

        # Document-specific controls
        showSvgButton = (result.delta is not None
                         and result.delta.new.path.lower().endswith(".svg")
                         and isImageFormatSupported("file.svg"))
        area.diffButtons.svgButton.setVisible(showSvgButton)

        # Set document
        if document is None:
            area.clearDocument(result.locator)

        elif isinstance(document, DiffDocument):
            assert result.delta is not None
            area.setDiffStackPage("text")
            area.diffView.replaceDocument(self.repo, result.delta, result.locator, document)

        elif isinstance(document, GitConflict):
            conflict = document
            area.setDiffStackPage("conflict")
            area.conflictView.displayConflict(conflict)

        elif isinstance(document, SpecialDiffError):
            area.setDiffStackPage("special")
            area.specialDiffView.displaySpecialDiffError(document)

        elif isinstance(document, ImageDelta):
            area.setDiffStackPage("special")
            area.specialDiffView.displayImageDelta(document)

        else:
            raise NotImplementedError(f"Can't display {type(document)}")

    @staticmethod
    def makeHeaderText(
            document: TAbstractDiffDocument | None,
            delta: GitDelta | None,
            locator: NavLocator
    ) -> str:
        if (document is None
                or delta is None
                or locator.context == NavContext.SPECIAL):
            return ""

        details = []

        if locator.context.isWorkdir():
            contextName = trtables.enum(locator.context).lower()
            details.append(contextName)
        elif locator.context == NavContext.COMMITTED:
            diffAB = locator.commitDiffAB()
            if diffAB:
                details.append(f"{shortHash(diffAB[0])}...{shortHash(diffAB[1])}")
            else:
                details.append(_p("at (specific commit)", "at {0}", shortHash(locator.commit)))

        if not delta.new.lfs.isTentative():
            if delta.new.lfs and not delta.old.lfs:
                details.append(tagify(_("LFS pointer [added]"), "<add>"))
            elif not delta.new.lfs and delta.old.lfs:
                details.append(tagify(_("LFS pointer [removed]"), "<del>"))
            elif delta.new.lfs and delta.old.lfs:
                details.append(_("LFS pointer changed"))
        elif delta.old.lfs:
            details[0] = _("unstaged changes to LFS object")

        # Compose final message
        parts = ["<html>", settings.prefs.addDelColorsStyleTag(), escape(locator.path)]
        if isinstance(document, DiffDocument):
            if document.pluses:
                parts.append(f" <b><add>+{document.pluses}</add></b>")
            if document.minuses:
                parts.append(f" <b><del>-{document.minuses}</del></b>")
        if details:
            suffix = ", ".join(details)
            parts.append(f" <span style='color: gray;'>({suffix})</span>")
        return "".join(parts)

    @classmethod
    def _jumpDelta(cls, task: RepoTask, delta: int):
        """
        Navigate back or forward in the RepoWidget's NavHistory.
        """

        assert delta in [-1, 1], "illegal delta value"

        rw = task.rw

        # Get starting point
        rw.saveFilePositions()
        start = rw.navLocator

        # Work on a copy of the history while we jump back/forward.
        history = rw.navHistory.copy()

        # Keep jumping back (or forward) in the history until the current
        # locator differs from the starting point, or until the history is
        # exhausted.
        while history.canGoDelta(delta):
            # Move back or forward in the history
            locator = history.navigateDelta(delta)

            # Keep going if same file comes up several times in a row
            if locator.isSimilarEnoughTo(start):
                continue

            # Do the jump. This may be a no-op if the locator is stale.
            yield from task.flowSubtask(cls, locator)

            # The jump was successful if the RepoWidget's locator
            # comes out similar enough to the one from the history.
            if rw.navLocator.isSimilarEnoughTo(locator):
                break

            # This point in history is stale, nuke it and keep going
            history.popCurrent()

        # Finalize history
        history.push(rw.navLocator)
        rw.navHistory = history
        rw.historyChanged.emit()


class JumpBack(RepoTask):
    def flow(self):
        yield from Jump._jumpDelta(self, -1)


class JumpForward(RepoTask):
    def flow(self):
        yield from Jump._jumpDelta(self, 1)


class JumpToUncommittedChanges(RepoTask):
    def flow(self):
        locator = NavLocator.inWorkdir()
        yield from self.flowSubtask(Jump, locator)


class JumpToHEAD(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoUnborn

    def flow(self):
        locator = NavLocator.inRef("HEAD")
        yield from self.flowSubtask(Jump, locator)


class RefreshRepo(RepoTask):
    @staticmethod
    def canKill_static(task: RepoTask):
        return task is None or isinstance(task, Jump | RefreshRepo)

    def canKill(self, task: RepoTask):
        return RefreshRepo.canKill_static(task)

    def onError(self, exc: Exception):
        # Don't refresh again if this task was interrupted by an error.
        # Note that the effects will be kept until the next refresh if this
        # task was killed by another task (this doesn't count as an error).
        self.epilog.effects = TaskEffects.Nothing

        super().onError(exc)

    def flow(self, effectFlags: TaskEffects = TaskEffects.DefaultRefresh, jumpTo: NavLocator = NavLocator.Empty):
        rw = self.rw
        repoModel = self.repoModel
        assert onAppThread()

        if effectFlags == TaskEffects.Nothing:
            return

        # Early out if repo has gone missing
        if not os.path.isdir(self.repo.path):
            raise RepoGoneError(self.repo.path)

        # Accumulate effect bits until task is complete or interrupted by an error
        self.epilog.effects |= effectFlags

        repoModel.workdirStale |= bool(effectFlags & TaskEffects.Workdir)

        initialLocator = rw.navLocator
        initialGraphScroll = rw.graphView.verticalScrollBar().value()
        restoringInitialLocator = jumpTo.context == NavContext.EMPTY
        wasExploringDetachedCommit = initialLocator.commit and initialLocator.commit not in repoModel.graph.commitRows

        jumpTo = jumpTo or initialLocator
        pNumUncommittedChanges = repoModel.numUncommittedChanges

        try:
            previousFileList = rw.diffArea.fileListByContext(initialLocator.context)
            previousFileList.backUpSelection()
        except ValueError:
            previousFileList = None

        refsChanged = False
        stashesChanged = False
        submodulesChanged = False
        remotesChanged = False
        upstreamsChanged = False
        homeBranchChanged = False

        if effectFlags & TaskEffects.Head:
            # Refresh the index. Useful in vanilla git mode: git may have touched
            # the index file during the task, so make libgit2 aware of it.
            repoModel.repo.refresh_index()

        if effectFlags & (TaskEffects.Head | TaskEffects.Workdir):
            submodulesChanged = repoModel.syncSubmodules()

        if effectFlags & (TaskEffects.Refs | TaskEffects.Remotes):
            remotesChanged = repoModel.syncRemotes()

        if effectFlags & (TaskEffects.Refs | TaskEffects.Upstreams):
            upstreamsChanged = repoModel.syncUpstreams()

        if effectFlags & (TaskEffects.Refs | TaskEffects.Remotes | TaskEffects.Head):
            # Refresh ref cache
            oldRefs = repoModel.refs
            oldHeadBranch = repoModel.homeBranch

            refsChanged = repoModel.syncRefs()
            refsChanged |= repoModel.syncMergeheads()
            stashesChanged = repoModel.syncStashes()
            homeBranchChanged = oldHeadBranch != repoModel.homeBranch

            # Load commits from changed refs only
            if refsChanged:
                self.syncTopOfGraph(oldRefs)

        # Refresh ahead/behind. Although Repository.ahead_behind exists, it's
        # slower than "git --for-each-ref" in complex graphs.
        # TODO: Would it be faster to calc this with our pre-cached graph?
        if refsChanged or upstreamsChanged:
            # TODO: This bit is duplicated with PrimeRepo
            with Benchmark("ahead-behind"):
                driver = yield from self.flowCallGit("for-each-ref", "--format=%(refname:short) %(upstream:track)", "refs/heads")
                repoModel.aheadBehind = dict(parseAheadBehind(driver.stdoutScrollback()))

        # Schedule a repaint of the entire GraphView if the refs changed
        if effectFlags & (TaskEffects.Head | TaskEffects.Refs):
            rw.graphView.viewport().update()

        # Refresh sidebar
        rw.sidebar.backUpSelection()
        anyChanges = refsChanged | stashesChanged | submodulesChanged | remotesChanged | homeBranchChanged | upstreamsChanged
        if anyChanges:
            with QSignalBlockerContext(rw.sidebar):
                rw.sidebar.refresh(repoModel)

        # Now jump to where we should be after the refresh
        assert rw.navLocator == initialLocator, "locator has changed"

        jumpToWorkdir = jumpTo.context.isWorkdir() or (jumpTo.context == NavContext.EMPTY and initialLocator.context.isWorkdir())

        if jumpToWorkdir:
            # Refresh workdir view on separate thread AFTER all the processing above
            if not jumpTo.context.isWorkdir():
                jumpTo = NavLocator(NavContext.WORKDIR)

            if effectFlags & TaskEffects.Workdir:
                newFlags = jumpTo.flags | NavFlags.ForceDiff | NavFlags.AllowWriteIndex
                jumpTo = jumpTo.replace(flags=newFlags)

        elif initialLocator and initialLocator.context == NavContext.COMMITTED:
            # After inserting/deleting rows in the commit log model,
            # the selected row may jump around. Try to restore the initial
            # locator to ensure the previously selected commit stays selected.
            rw.graphView.verticalScrollBar().setValue(initialGraphScroll)
            if (jumpTo == initialLocator
                    and jumpTo.commit not in repoModel.graph.commitRows
                    and not wasExploringDetachedCommit):
                # We were looking at a commit that is not in the graph anymore.
                # Probably refreshing after amending. Jump to HEAD commit.
                jumpTo = NavLocator.inCommit(repoModel.headCommitId)

        # Jump
        yield from self.flowSubtask(Jump, jumpTo)

        # Try to restore sidebar selection
        if restoringInitialLocator:
            rw.sidebar.restoreSelectionBackup()
        else:
            rw.sidebar.clearSelectionBackup()

        # Try to restore path selection
        if previousFileList is None:
            pass
        elif restoringInitialLocator:
            previousFileList.restoreSelectionBackup()
        else:
            previousFileList.clearSelectionBackup()

        # If workdir is still stale (refreshing without explicitly looking at the workdir), refresh it after everything else.
        # This is done last so that it doesn't impede on responsivity when the user isn't explicitly looking at the workdir.
        if repoModel.workdirStale:
            assert not jumpToWorkdir, "jumping to workdir should have refreshed the workdir!"
            yield from loadWorkdir(self, jumpTo.hasFlags(NavFlags.AllowWriteIndex))

        # Update number of staged changes in sidebar and graph
        if repoModel.numUncommittedChanges != pNumUncommittedChanges:
            rw.refreshNumUncommittedChanges()

        # Refresh window title and state banner.
        # Do this last because it requires the index to be fresh (updated by LoadWorkdir)
        rw.refreshWindowTitle()
        rw.refreshBanner()

        # Finally, clear out the effect bits that were accumulated by
        # RepoWidget.refreshRepo. Do this *last* because we want to keep the
        # effect bits for the next RefreshRepo task if this one was interrupted
        # by another task. Note that interrupting by an error will flush the bits
        # so we don't get stuck in a refresh loop.
        self.epilog.effects = TaskEffects.Nothing

        if anyChanges:
            logger.debug(f"Changes detected on refresh: Ref={int(refsChanged)} Sta={int(stashesChanged)} "
                         f"Sub={int(submodulesChanged)} Rem={int(remotesChanged)} Ups={int(upstreamsChanged)}")

    def syncTopOfGraph(self, oldRefs: dict[str, Oid]):
        repoModel = self.repoModel
        graphView = self.rw.graphView
        clModel = graphView.clModel
        clFilter = graphView.clFilter

        # Make sure we're on the UI thread.
        # We don't want GraphView to try to read an incomplete state while repainting.
        assert onAppThread()

        # Update our graph model
        gsl = repoModel.syncTopOfGraph(oldRefs)

        with QSignalBlockerContext(graphView):
            # Sync top of graphview
            clModel.resetCommitSequence(gsl.numRowsRemoved, gsl.numRowsAdded)

            # Hidden commits may have changed in RepoModel.syncTopOfGraph!
            # If new commits are part of a hidden branch, we must invalidate CommitLogFilter.
            clFilter.updateHiddenCommits()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette import trtables
from gitfourchette.forms.ui_signatureform import Ui_SignatureForm
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class SignatureOverride(enum.IntEnum):
    Nothing = 0
    Author = 1
    Committer = 2
    Both = 3


class SignatureForm(QWidget):
    signatureChanged = Signal()

    @staticmethod
    def validateInput(item: str) -> str:
        try:
            validate_signature_item(item)
            return ""
        except NameValidationError as exc:
            return trtables.enum(exc.rule)

    def __init__(self, parent):
        super().__init__(parent)
        self.ui = Ui_SignatureForm()
        self.ui.setupUi(self)
        try:
            self.ui.timeEdit.setTimeZone(QTimeZone.systemTimeZone())
        except AttributeError:  # pragma: no cover - TODO: Remove once we can drop compatibility with Qt <6.7.
            self.ui.timeEdit.setTimeSpec(Qt.TimeSpec.LocalTime)
        self.primeTimeOffsetComboBox()

        self.ui.replaceComboBox.currentIndexChanged.connect(self.signatureChanged)
        self.ui.nameEdit.textChanged.connect(self.signatureChanged)
        self.ui.emailEdit.textChanged.connect(self.signatureChanged)
        self.ui.timeEdit.timeChanged.connect(self.signatureChanged)
        self.ui.offsetEdit.currentIndexChanged.connect(self.signatureChanged)
        self.ui.nowButton.clicked.connect(self.setDateTimeNow)

    def setSignature(self, signature: Signature):
        qdt = signatureQDateTime(signature)
        with QSignalBlockerContext(self):
            self.ui.nameEdit.setText(signature.name)
            self.ui.emailEdit.setText(signature.email)
            self.ui.timeEdit.setDateTime(qdt)
            self.setTimeOffset(signature.offset)
        self.signatureChanged.emit()

    def primeTimeOffsetComboBox(self):
        decimalCodedOffsets = [
            -11_00, -10_00, -9_00, -9_30, -8_00, -7_00,
            -6_00, -5_00, -4_00, -3_30, -3_00, -2_30, -2_00, -1_00,
            +0, +1_00, +2_00, +3_00, +3_30, +4_00, +4_30, +5_00, +5_30, +5_45,
            +6_00, +6_30, +7_00, +8_00, +9_00, +9_30, +10_00, +10_30, +11_00,
            +12_00, +13_00,
        ]

        for t in decimalCodedOffsets:
            minutes = (abs(t) // 100) * 60 + (abs(t) % 100)
            if t < 0:
                minutes = -minutes
            self.ui.offsetEdit.addItem(formatTimeOffset(minutes), minutes)

        # Prime QComboBox for local timezone (if it's missing from the list above)
        self.setDateTimeNow()

    def setTimeOffset(self, minutes):
        comboBox = self.ui.offsetEdit
        i = comboBox.findData(minutes)

        # If an obscure offset is missing from our QComboBox, insert it
        if i < 0:
            # Find where to insert it
            for i in range(comboBox.count()):
                existingOffset = comboBox.itemData(i)
                if existingOffset >= minutes:
                    break
            else:
                i = comboBox.count()
            comboBox.insertItem(i, formatTimeOffset(minutes), minutes)

        assert i >= 0
        comboBox.setCurrentIndex(i)

    def setDateTimeNow(self):
        now = QDateTime.currentDateTime()
        self.ui.timeEdit.setDateTime(now)
        self.setTimeOffset(now.offsetFromUtc() // 60)

    def installValidator(self, validator: ValidatorMultiplexer):
        validator.connectInput(self.ui.nameEdit, SignatureForm.validateInput)
        validator.connectInput(self.ui.emailEdit, SignatureForm.validateInput)

    def getSignature(self) -> Signature | None:
        qdt = self.ui.timeEdit.dateTime()
        try:
            return Signature(
                name=self.ui.nameEdit.text(),
                email=self.ui.emailEdit.text(),
                time=qdt.toSecsSinceEpoch(),
                offset=self.ui.offsetEdit.currentData(),
            )
        except ValueError:
            return None

    def replaceWhat(self):
        index = self.ui.replaceComboBox.currentIndex()
        return SignatureOverride(index + 1)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *
from gitfourchette.toolbox import *


class BlameBusySpinner(QBusySpinner):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setFixedSize(96, 96)

        self.delayStart = QTimer(self)
        self.delayStart.timeout.connect(self.show)
        self.delayStart.setInterval(200)
        self.delayStart.setSingleShot(True)

        self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

        self.reposition()
        self.hide()

    def reposition(self):
        self.move((self.parentWidget().width() - self.width()) // 2, 64)

    def start(self):
        if not self.delayStart.isActive():
            self.delayStart.start()

    def stop(self):
        self.delayStart.stop()
        self.hide()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *
from gitfourchette.toolbox import *


class QStatusBar2(QStatusBar):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setObjectName("QStatusBar2")

        self.memoryIndicator = MemoryIndicator(self)

        self.setSizeGripEnabled(False)
        self.addPermanentWidget(self.memoryIndicator)
        # macOS: must reset stylesheet after addPermanentWidget for no-border thickness thing to take effect
        self.memoryIndicator.setStyleSheet(self.memoryIndicator.styleSheet())

        self.busyMessageDelayer = QTimer(self)
        self.busyMessageDelayer.setSingleShot(True)
        self.busyMessageDelayer.setInterval(20)
        self.busyMessageDelayer.timeout.connect(self.commitBusyMessage)

        self.busyWidget = QWidget(self)
        self.busySpinner = QBusySpinner(self.busyWidget)
        self.busyLabel = QElidedLabel(self.busyWidget)
        # Emojis such as the lightbulb may increase the label's height
        self.busyWidget.setMaximumHeight(self.fontMetrics().height())

        layout = QHBoxLayout(self.busyWidget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.busySpinner)
        layout.addWidget(self.busyLabel, 1)

        self.busyWidget.setVisible(False)
        self.isBusyMessageSet = False

    def showMessage(self, text: str, msecs=0):
        if self.busyMessageDelayer.isActive():
            # Commit the busy message now and kill the delayer.
            # It'll be immediately overridden by the temporary message.
            # We don't want it to happen the other way around.
            self.commitBusyMessage()

        super().showMessage(text, msecs)

    def showBusyMessage(self, text: str):
        self.busyLabel.setText(text)
        if not self.isBusyMessageSet and not self.busyMessageDelayer.isActive():
            self.busyMessageDelayer.start()

    def commitBusyMessage(self):
        self.busyMessageDelayer.stop()

        if self.isBusyMessageSet:
            return

        # Replace permanent status message with our busyWidget
        # which includes the spinner and the busy message.
        self.isBusyMessageSet = True
        self.busySpinner.setVisible(True)
        self.busyWidget.setVisible(True)
        self.insertPermanentWidget(0, self.busyWidget, 1)

    def clearMessage(self):
        self.busyMessageDelayer.stop()

        if self.isBusyMessageSet:
            self.isBusyMessageSet = False
            self.removeWidget(self.busyWidget)
            self.busyWidget.setVisible(False)

        super().clearMessage()

    def enableMemoryIndicator(self, show: bool = False):
        self.memoryIndicator.setVisible(show)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import dataclasses
import enum
import fnmatch
import hashlib
import os
from pathlib import Path

from pygit2.enums import AttrCheck

from gitfourchette.gitdriver.lfspointer import LfsPointer, LfsPointerState, LfsPointerMagicBytes, LfsPointerPattern, LfsObjectCacheMissingError
from gitfourchette.porcelain import FileMode, ObjectType, Repo, Oid, id7, pygit2_version_at_least

HexHash0000 = "0" * 40
HexHashFFFF = "f" * 40

HexHashEmptyBlob = "e69de29bb2d1d6434b8b29ae775ad8c2e48c5391"  # hashlib.sha1(b'blob 0\0').hexdigest()
""" The SHA-1 hash of an empty Git blob. """

_NoDiskStat = (-1, -1)

_UnknownLfsPointer = LfsPointer(LfsPointerState.Unknown)
_NoLfsPointer = LfsPointer(LfsPointerState.NoPointer)

_sha1ToSha256Equivalent: dict[str, str] = {}
"""
Map SHA-1 blob hashes to SHA-256 equivalents for quick comparison of regular
git blobs (SHA-1) to LFS pointers (SHA-256).
"""


class GitDeltaSource(enum.IntEnum):
    Unknown = 0
    Dirty = 1
    Index = 2
    Commit = 3

    def isWorkdir(self) -> bool:
        return self == GitDeltaSource.Dirty or self == GitDeltaSource.Index


@dataclasses.dataclass
class GitDeltaFile:
    path: str = ""
    """
    Workdir-relative path to the file.
    Directory separators are forward slashes, even on Windows.
    """

    id: str = HexHash0000
    """
    SHA-1 hash of the blob, in hexadecimal format (40 characters).
    All 0s means there's no blob (i.e. file deleted).
    All Fs means the hash hasn't been computed.
    """

    mode: FileMode = FileMode.UNREADABLE
    """
    Mode of the file as far as Git is concerned. Be careful: Git records
    simplified modes that may not accurately reflect a file's actual mode
    in the filesystem.
    """

    source: GitDeltaSource = GitDeltaSource.Unknown
    """
    Where this GitDeltaFile was sampled from: a commit, the index, or an
    unstaged change or untracked file in the workdir.
    """

    sourceCommit: Oid | None = None
    """
    Only valid when source == GitDeltaSource.Commit.
    """

    diskStat: tuple[int, int] = _NoDiskStat
    """
    Filled in for unstaged files only. Allows quick comparison of GitDeltaFiles
    taken at two points in time for the same unstaged file. Internally, this is
    a snapshot of a subset of the file's status on disk (st_mtime_ns, st_size).
    """

    _data: bytes | None = dataclasses.field(default=None, compare=False)
    """
    Cached file contents. Not used in object comparisons.
    None means that the file hasn't been cached yet (isDataValid() == False).
    """

    lfs: LfsPointer = dataclasses.field(default=_UnknownLfsPointer, compare=False)
    """
    Cached LFS object information extracted from the LFS pointer (if any).
    """

    SupportsFastSizeBallpark = pygit2_version_at_least("1.19.2", False, "fast size ballpark")

    def __post_init__(self):
        assert self.id.isnumeric() or self.id.islower()
        if self.isId0():
            self._data = b""

    def __bool__(self) -> bool:
        return not self.isId0()

    def isId0(self) -> bool:
        return self.id == HexHash0000

    def isEmptyBlob(self) -> bool:
        return self.id == HexHashEmptyBlob

    def isIdValid(self) -> bool:
        return self.id != HexHashFFFF

    def isDataValid(self) -> bool:
        return self._data is not None

    def hasDiskStat(self) -> bool:
        return self.diskStat != _NoDiskStat

    def isBlob(self) -> bool:
        return self.mode & FileMode.BLOB == FileMode.BLOB

    def read(self, repo: Repo, maxSize: int = -1) -> bytes:
        if self._data is not None:
            # Data already cached
            pass

        elif self.lfs.state == LfsPointerState.UnstagedTentative:
            # Would be an LFS file once staged, read it direct from the workdir
            assert self.lfs.size < 0
            self._data = repo.apply_filters_to_workdir(self.path)
            self.lfs = dataclasses.replace(self.lfs, size=len(self._data))

        elif self.lfs.state == LfsPointerState.Valid:
            # LFS pointer resolved, load data from LFS object db
            try:
                self._data = Path(self.lfs.objectPath).read_bytes()
            except FileNotFoundError as fnf:
                raise LfsObjectCacheMissingError(self.lfs) from fnf
            assert self.lfs.size == len(self._data), "LFS object size mismatch"

        else:
            try:
                # Load blob from standard git object database
                if not self.isIdValid():  # unknown hash (FFFFFFF...)
                    raise KeyError()

                # Honor size heuristic, if any
                if 0 <= maxSize < self.sizeBallpark(repo):
                    raise OverflowError("blob size exceeds limit")

                self._data = repo.peel_blob(self.id).data

            except KeyError:
                # Blob ID isn't in the database. Typically, that means
                # it's an unstaged file. Read it from the workdir.
                assert self.source == GitDeltaSource.Dirty, f"expecting untracked/unstaged, got {self.source}"
                self._data = repo.apply_filters_to_workdir(self.path)

        assert self.isDataValid(), "data should be valid here"
        return self._data

    def stat(self, repo: Repo) -> tuple[int, int]:
        diskStat = _NoDiskStat
        absPath = repo.in_workdir(self.path)
        try:
            stat = os.lstat(absPath)
            diskStat = (stat.st_mtime_ns, stat.st_size)
        except OSError:
            pass
        return diskStat

    def sizeBallpark(self, repo: Repo) -> int:
        if self.isId0():
            return 0

        if self.lfs.size >= 0:
            return self.lfs.size

        if self.isIdValid() and self.SupportsFastSizeBallpark:
            try:
                blobType, blobSize = repo.odb.read_header(self.id)
                assert blobType == ObjectType.BLOB
                return blobSize
            except KeyError:
                # Fall back to workdir stat
                pass

        if self.isIdValid() and not self.SupportsFastSizeBallpark:  # pragma: no cover (Odb.read_header missing in old versions of pygit2)
            # The mere act of looking up a blob can be much slower than Odb.read_header for large blobs!
            # TODO: Remove once we can stop supporting pygit2 <= 1.19.1
            try:
                return repo.peel_blob(self.id).size
            except KeyError:
                pass

        assert self.source.isWorkdir(), "can't estimate non-workdir blob size without a blob ID"

        _, size = self.stat(repo)
        return size

    def cacheLfsPointer(self, repo: Repo, checkOrKnownValue: AttrCheck | str):
        # No-op if LFS pointer already resolved
        if self.lfs.state != LfsPointerState.Unknown:
            return

        # No LFS pointer if the file doesn't exist at this point
        if self.isId0():
            self.lfs = _NoLfsPointer
            return

        # Get filter attribute for this file
        attr: str | bool | None
        if isinstance(checkOrKnownValue, str):
            attr = checkOrKnownValue
        else:
            check = checkOrKnownValue
            attr = repo.get_attr(self.path, "filter", check, self.sourceCommit)

        # File must have 'lfs' filter attribute past this point
        if attr != "lfs":
            self.lfs = _NoLfsPointer
            return

        # Unstaged: Force read data from wd
        if self.source == GitDeltaSource.Dirty:
            assert self.sourceCommit is None
            objectPath = repo.in_workdir(self.path)
            self.lfs = LfsPointer(LfsPointerState.UnstagedTentative, objectPath=objectPath)
            return

        # Analyze the non-LFS blob to make sure it's really an LFS pointer.
        try:
            data = self.read(repo, maxSize=256)
        except OverflowError:
            # Don't bother loading large blobs; unlikely to be a valid pointer
            data = b""

        if not data.startswith(LfsPointerMagicBytes):
            self.lfs = _NoLfsPointer
            return

        text = data.decode("utf-8", errors="replace")
        match = LfsPointerPattern.match(text)
        sha = match.group(1)
        size = int(match.group(2))
        objectPath = repo.in_gitdir(f"lfs/objects/{sha[:2]}/{sha[2:4]}/{sha}")
        self.lfs = LfsPointer(LfsPointerState.Valid, sha, size, objectPath, text)

        # Invalidate data so that next read() uses LFS data
        self._data = None

    def blobSha256(self, repo) -> str:
        try:
            sha256 = _sha1ToSha256Equivalent[self.id]
        except KeyError:
            blobContents = self.read(repo)
            sha256 = hashlib.sha256(blobContents).hexdigest()
            _sha1ToSha256Equivalent[self.id] = sha256
        return sha256

    def matchPathspec(self, pathspec: str) -> bool:
        # I've looked into adding git_pathspec_matches_path to pygit2, but it
        # doesn't support magic signatures like :(icase), so it wouldn't be much
        # of an improvement over plain fnmatch. (4/2026)

        assert pathspec == pathspec.lower()
        return fnmatch.fnmatch(self.path.lower(), pathspec)

    def __repr__(self) -> str:
        return f"({self.path},{id7(self.id)},{self.mode:o})"


GitDeltaFile_Empty = GitDeltaFile()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import os
import time
from collections.abc import Callable
from typing import ClassVar

BENCHMARK_LOGGING_LEVEL = 5

logger = logging.getLogger(__name__)
logging.addLevelName(BENCHMARK_LOGGING_LEVEL, "BENCHMARK")

try:
    import psutil
except ModuleNotFoundError:
    psutil = None


def getRSS():
    if psutil:
        return psutil.Process(os.getpid()).memory_info().rss
    else:
        return 0


class Benchmark:
    """ Context manager that reports how long a piece of code takes to run. """

    nesting: ClassVar[list[str]] = []

    def __init__(self, name: str):
        self.name = name
        self.phase = ""
        self.startTime = 0.0
        self.startBytes = 0

    def enter(self, phase=""):
        if self.startTime:
            self.exit()

        Benchmark.nesting.append(self.name)
        self.startBytes = getRSS()
        self.startTime = time.perf_counter()
        self.phase = phase

    def exit(self, exc_type=None):
        ms = 1000 * self.elapsed()
        kb = (getRSS() - self.startBytes) // 1024

        description = "/".join(Benchmark.nesting)
        if self.phase:
            description += f" ({self.phase})"
        if exc_type:
            description += f" (EXCEPTION RAISED! {exc_type.__name__})"
        logger.log(BENCHMARK_LOGGING_LEVEL, f"{ms:8.1f} ms {kb:6,d}K {description}")

        Benchmark.nesting.pop()
        self.startTime = 0.0
        self.phase = ""

    def elapsed(self):
        return time.perf_counter() - self.startTime

    def __enter__(self):
        self.enter()
        return self

    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        self.exit(exc_type)


def benchmark(func: Callable) -> Callable:
    """ Function decorator that reports how long the function takes to run. """
    def wrapper(*args, **kwargs) -> Callable:
        with Benchmark(func.__qualname__):
            return func(*args, **kwargs)
    return wrapper



# Form implementation generated from reading ui file 'signatureform.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_SignatureForm(object):
    def setupUi(self, SignatureForm):
        SignatureForm.setObjectName("SignatureForm")
        SignatureForm.resize(340, 96)
        self.formLayout = QFormLayout(SignatureForm)
        self.formLayout.setContentsMargins(0, 0, 0, 0)
        self.formLayout.setObjectName("formLayout")
        self.replaceLabel = QLabel(parent=SignatureForm)
        self.replaceLabel.setObjectName("replaceLabel")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.replaceLabel)
        self.nameLabel = QLabel(parent=SignatureForm)
        self.nameLabel.setObjectName("nameLabel")
        self.formLayout.setWidget(1, QFormLayout.ItemRole.LabelRole, self.nameLabel)
        self.identityLayout = QHBoxLayout()
        self.identityLayout.setObjectName("identityLayout")
        self.nameEdit = QLineEdit(parent=SignatureForm)
        self.nameEdit.setObjectName("nameEdit")
        self.identityLayout.addWidget(self.nameEdit)
        self.emailEdit = QLineEdit(parent=SignatureForm)
        self.emailEdit.setObjectName("emailEdit")
        self.identityLayout.addWidget(self.emailEdit)
        self.formLayout.setLayout(1, QFormLayout.ItemRole.FieldRole, self.identityLayout)
        self.timeLabel = QLabel(parent=SignatureForm)
        self.timeLabel.setObjectName("timeLabel")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.LabelRole, self.timeLabel)
        self.timeLayout = QHBoxLayout()
        self.timeLayout.setObjectName("timeLayout")
        self.timeEdit = QDateTimeEdit(parent=SignatureForm)
        self.timeEdit.setCurrentSection(QDateTimeEdit.Section.YearSection)
        self.timeEdit.setDisplayFormat("yyyy-MM-dd HH:mm:ss")
        self.timeEdit.setCalendarPopup(True)
        self.timeEdit.setObjectName("timeEdit")
        self.timeLayout.addWidget(self.timeEdit)
        self.offsetEdit = QComboBox(parent=SignatureForm)
        self.offsetEdit.setEditable(False)
        self.offsetEdit.setObjectName("offsetEdit")
        self.timeLayout.addWidget(self.offsetEdit)
        self.nowButton = QToolButton(parent=SignatureForm)
        self.nowButton.setObjectName("nowButton")
        self.timeLayout.addWidget(self.nowButton)
        self.formLayout.setLayout(2, QFormLayout.ItemRole.FieldRole, self.timeLayout)
        self.replaceComboBox = QComboBox(parent=SignatureForm)
        self.replaceComboBox.setObjectName("replaceComboBox")
        self.replaceComboBox.addItem("")
        self.replaceComboBox.addItem("")
        self.replaceComboBox.addItem("")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.replaceComboBox)
        self.replaceLabel.setBuddy(self.replaceComboBox)
        self.nameLabel.setBuddy(self.nameEdit)
        self.timeLabel.setBuddy(self.timeEdit)

        self.retranslateUi(SignatureForm)
        SignatureForm.setTabOrder(self.replaceComboBox, self.nameEdit)
        SignatureForm.setTabOrder(self.nameEdit, self.emailEdit)
        SignatureForm.setTabOrder(self.emailEdit, self.timeEdit)
        SignatureForm.setTabOrder(self.timeEdit, self.offsetEdit)
        SignatureForm.setTabOrder(self.offsetEdit, self.nowButton)

    def retranslateUi(self, SignatureForm):
        SignatureForm.setWindowTitle(_p("SignatureForm", "Signature"))
        self.replaceLabel.setText(_p("SignatureForm", "&Override:"))
        self.nameLabel.setText(_p("SignatureForm", "&Identity:"))
        self.nameEdit.setToolTip(_p("SignatureForm", "Name"))
        self.nameEdit.setPlaceholderText(_p("SignatureForm", "Name"))
        self.emailEdit.setToolTip(_p("SignatureForm", "Email"))
        self.emailEdit.setPlaceholderText(_p("SignatureForm", "Email"))
        self.timeLabel.setText(_p("SignatureForm", "&Time:"))
        self.offsetEdit.setToolTip(_p("SignatureForm", "Offset from UTC"))
        self.nowButton.setToolTip(_p("SignatureForm", "Set to current local time"))
        self.nowButton.setText(_p("SignatureForm", "Now"))
        self.replaceComboBox.setToolTip(_p("SignatureForm", "<p>Every commit records information about an author and a committer. They can be the same person.</p>\n"
"       <p>Select which person you want to override.</p>"))
        self.replaceComboBox.setItemText(0, _p("SignatureForm", "Author"))
        self.replaceComboBox.setItemText(1, _p("SignatureForm", "Committer"))
        self.replaceComboBox.setItemText(2, _p("SignatureForm", "Both"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from collections.abc import Callable

from gitfourchette.gitdriver import *
from gitfourchette.porcelain import MultiFileError
from gitfourchette.tasks import RepoTask
from gitfourchette.tasks.repotask import showMultiFileErrorMessage
from gitfourchette.toolbox import tagify


class FileBatchTask(RepoTask):
    UnitFunc = Callable[[RepoTask, GitDelta], RepoTask.Flow[None]]

    def flow(
            self,
            deltaList: list[GitDelta],
            unitFunc: UnitFunc,
            title: str,
            prompt: str,
            threshold=3,
    ):
        assert "#" in prompt

        numFiles = len(deltaList)

        if numFiles > threshold:
            prompt = tagify(prompt.replace("#", str(numFiles)), "<b>")
            detailList = [d.new.path for d in deltaList]
            yield from self.flowConfirm(title, prompt, detailList=detailList)

        errors = MultiFileError()

        for delta in deltaList:
            try:
                yield from unitFunc(self, delta)
                errors.add_file_success()
            except (OSError,  # typically FileNotFoundError
                    LfsObjectCacheMissingError
                    ) as exc:
                errors.add_file_error(delta.new.path, exc)

        if errors:
            showMultiFileErrorMessage(self.parentWidget(), errors, title)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import functools
import sys
import threading
from collections.abc import Callable


def calledFromQThread(f: Callable):  # pragma: no cover
    """
    Add this decorator to functions that are called from within a QThread
    so pytest-cov can trace them correctly.
    """
    # Adapted from https://github.com/nedbat/coveragepy/issues/686#issuecomment-634932753

    if "coverage" not in sys.modules:
        return f

    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        sys.settrace(threading._trace_hook)  # type: ignore[attr-defined]
        return f(*args, **kwargs)

    return wrapped



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations  # TODO: Remove once we can drop support for Python <= 3.13

from gitfourchette.qt import *
from gitfourchette.search.searchprovider import SearchProvider

if TYPE_CHECKING:
    from gitfourchette.codeview.codeview import CodeView


class CodeSearch(SearchProvider):
    _buddy: CodeView

    def __init__(self, parent: CodeView):
        super().__init__(parent)
        self._buddy = parent

    def invalidate(self):
        super().invalidate()
        self._buddy.highlighter.setSearchTerm("")

    def _termChanged(self):
        numOccurrences = self._buddy.highlighter.setSearchTerm(self._term)

        if not self._term:  # Empty
            self.setStatus(SearchProvider.TermStatus.Unknown)
        elif numOccurrences == 0:
            self._enshrineBadTerm()
        else:
            self.setStatus(SearchProvider.TermStatus.Good)

    def isCurrentMatch(self) -> bool:
        assert self.isGoodAndNonEmpty()
        cursor: QTextCursor = self._buddy.textCursor()
        return cursor.selectedText().lower().startswith(self._term)

    def jump(self, forward: bool):
        assert self.isGoodAndNonEmpty()

        startCursor = self._buddy.textCursor()
        document = self._buddy.document()
        findFlags = [] if forward else [QTextDocument.FindFlag.FindBackward]

        for _wrap in range(2):  # Wrap around at most once
            newCursor = document.find(self._term, startCursor, *findFlags)

            if newCursor and not newCursor.isNull():  # extra isNull check needed for PyQt5 & PyQt6
                self._buddy.setTextCursor(newCursor)
                assert self.isCurrentMatch()
                return

            # Wrap
            startCursor.movePosition(QTextCursor.MoveOperation.Start if forward else QTextCursor.MoveOperation.End)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import dataclasses
import enum
from collections.abc import Callable, Iterable
from typing import ClassVar

from gitfourchette.qt import *
from gitfourchette.toolbox.iconbank import stockIcon
from gitfourchette.toolbox.qtutils import MultiShortcut, ShortcutKeys, appendShortcutToToolTipText


@dataclasses.dataclass
class ActionDef:
    """
    Build QMenus quickly with a list of ActionDefs.
    """

    SEPARATOR: ClassVar[ActionDef]
    SPACER: ClassVar[ActionDef]

    class Kind(enum.IntEnum):
        Action = enum.auto()
        Section = enum.auto()
        Separator = enum.auto()
        Spacer = enum.auto()

    caption: str = ""
    callback: SignalInstance | Callable | None = None
    icon: str = ""
    checkState: int = 0
    radioGroup: str = ""
    enabled: bool = True
    submenu: list[ActionDef] | QMenu = dataclasses.field(default_factory=list)
    shortcuts: MultiShortcut | ShortcutKeys = ""
    tip: str = ""
    objectName: str = ""
    menuRole: QAction.MenuRole = QAction.MenuRole.NoRole
    kind: Kind = Kind.Action

    def replace(self, **changes):
        return dataclasses.replace(self, **changes)

    def toQAction(self, parent: QWidget) -> QAction:
        action = QAction(self.caption, parent=parent)

        if self.objectName:
            action.setObjectName(self.objectName)

        if self.callback is None:
            pass
        elif isinstance(self.callback, SignalInstance):
            action.triggered.connect(self.callback)
        else:
            cb = self.callback
            assert callable(cb)
            action.triggered.connect(lambda: cb())

        if self.icon:
            action.setIcon(stockIcon(self.icon))

        if self.checkState != 0:
            action.setCheckable(True)
            action.setChecked(self.checkState == 1)

        if self.tip:
            tip = self.tip
            if self.shortcuts and not isinstance(parent, QMenu):
                if isinstance(self.shortcuts, list):
                    tip = appendShortcutToToolTipText(tip, self.shortcuts[0])
                else:
                    tip = appendShortcutToToolTipText(tip, self.shortcuts)
            if "<" not in tip and len(tip) > 40:  # wrap long tooltips
                tip = f"<p>{tip}</p>"
            action.setToolTip(tip)

        # Enforce menu role (including NoRole) to prevent Qt/macOS from moving
        # items that start with the localized word for "Settings" to the
        # application menu (e.g. "Repository Settings" in French).
        action.setMenuRole(self.menuRole)

        action.setEnabled(bool(self.enabled))

        if self.shortcuts:
            if isinstance(self.shortcuts, list):
                action.setShortcuts(self.shortcuts)
            else:
                action.setShortcut(self.shortcuts)

        if self.submenu:
            assert isinstance(self.submenu, QMenu), "ActionDef.toQAction cannot be used for submenus given as a list"
            action.setMenu(self.submenu)

        return action

    def makeSubmenu(self, parent: QMenu) -> QMenu:
        assert self.submenu
        assert not isinstance(self.submenu, QMenu)

        submenu = ActionDef.makeQMenu(parent=parent, actionDefs=self.submenu)
        submenu.setObjectName(f"{parent.objectName()}_ActionDefSubmenu" if parent else "ActionDefSubmenu")
        submenu.setTitle(self.caption)
        submenu.setEnabled(self.enabled)
        if self.icon:
            submenu.setIcon(stockIcon(self.icon))

        return submenu

    @staticmethod
    def addToQMenu(menu: QMenu, *actionDefs: ActionDef | QAction):
        radioGroups: dict[str, QActionGroup] = {}

        for item in actionDefs:
            if isinstance(item, QAction):
                item.setParent(menu)  # reparent it
                menu.addAction(item)
            elif item.kind == ActionDef.Kind.Separator:
                menu.addSeparator()
            elif item.kind == ActionDef.Kind.Section:
                ActionDef.addSectionToQMenu(menu, item.caption)
            elif item.submenu and not isinstance(item.submenu, QMenu):
                submenu = item.makeSubmenu(parent=menu)
                menu.addMenu(submenu)
            elif item.kind == ActionDef.Kind.Action:
                action = item.toQAction(parent=menu)
                menu.addAction(action)

                groupKey = item.radioGroup
                if groupKey:
                    try:
                        group = radioGroups[groupKey]
                    except KeyError:
                        group = QActionGroup(menu)
                        group.setObjectName(f"ActionDefGroup{groupKey}")
                        group.setExclusive(True)
                        radioGroups[groupKey] = group
                    group.addAction(action)
            else:
                raise NotImplementedError(f"Unsupported ActionDef kind in menu: {item.kind}")

    @staticmethod
    def addToQToolBar(toolbar: QToolBar, *actionDefs: ActionDef | QAction):
        for item in actionDefs:
            if isinstance(item, QAction):
                action: QAction = item
                action.setParent(toolbar)  # reparent it
                action.setShortcut("")  # clear shortcut for toolbar
                toolbar.addAction(action)
            elif item.kind == ActionDef.Kind.Separator:
                toolbar.addSeparator()
            elif item.kind == ActionDef.Kind.Spacer:
                spacer = QWidget()
                spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
                toolbar.addWidget(spacer)
            elif item.submenu:
                raise NotImplementedError("Cannot add ActionDef submenus to QToolbar")
            elif item.kind == ActionDef.Kind.Action:
                assert not item.radioGroup
                action = item.toQAction(parent=toolbar)
                action.setShortcut("")  # clear shortcut for toolbar
                toolbar.addAction(action)
            else:
                raise NotImplementedError(f"Unsupported ActionDef kind in toolbar: {item.kind}")

    @staticmethod
    def makeQMenu(
            parent: QWidget,
            actionDefs: Iterable[ActionDef | QAction],
    ) -> QMenu:
        menu = QMenu(parent)
        menu.setObjectName(f"{parent.objectName()}_ActionDefMenu" if parent else "ActionDefMenu")
        ActionDef.addToQMenu(menu, *actionDefs)
        return menu

    @staticmethod
    def addSectionToQMenu(menu: QMenu, text: str):
        label = QLabel(text)
        label.setProperty("class", "ActionDefSection")
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        widgetAction = QWidgetAction(menu)
        widgetAction.setDefaultWidget(label)
        widgetAction.setEnabled(False)
        menu.addAction(widgetAction)
        menu.addSeparator()
        return widgetAction


ActionDef.SEPARATOR = ActionDef(kind=ActionDef.Kind.Separator)
ActionDef.SPACER = ActionDef(kind=ActionDef.Kind.Spacer)



# Form implementation generated from reading ui file 'newtagdialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_NewTagDialog(object):
    def setupUi(self, NewTagDialog):
        NewTagDialog.setObjectName("NewTagDialog")
        NewTagDialog.setWindowModality(Qt.WindowModality.WindowModal)
        NewTagDialog.resize(405, 303)
        self.formLayout = QFormLayout(NewTagDialog)
        self.formLayout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.label = QLabel(parent=NewTagDialog)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.LabelRole, self.label)
        self.nameEdit = QLineEdit(parent=NewTagDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.nameEdit.sizePolicy().hasHeightForWidth())
        self.nameEdit.setSizePolicy(sizePolicy)
        self.nameEdit.setObjectName("nameEdit")
        self.formLayout.setWidget(0, QFormLayout.ItemRole.FieldRole, self.nameEdit)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.pushCheckBox = QCheckBox(parent=NewTagDialog)
        self.pushCheckBox.setObjectName("pushCheckBox")
        self.horizontalLayout.addWidget(self.pushCheckBox)
        self.remoteComboBox = QComboBox(parent=NewTagDialog)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.remoteComboBox.sizePolicy().hasHeightForWidth())
        self.remoteComboBox.setSizePolicy(sizePolicy)
        self.remoteComboBox.setObjectName("remoteComboBox")
        self.horizontalLayout.addWidget(self.remoteComboBox)
        self.formLayout.setLayout(1, QFormLayout.ItemRole.FieldRole, self.horizontalLayout)
        self.forceCheckBox = QCheckBox(parent=NewTagDialog)
        self.forceCheckBox.setObjectName("forceCheckBox")
        self.formLayout.setWidget(2, QFormLayout.ItemRole.FieldRole, self.forceCheckBox)
        self.buttonBox = QDialogButtonBox(parent=NewTagDialog)
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.formLayout.setWidget(3, QFormLayout.ItemRole.SpanningRole, self.buttonBox)
        self.label.setBuddy(self.nameEdit)

        self.retranslateUi(NewTagDialog)
        self.buttonBox.accepted.connect(NewTagDialog.accept)
        self.buttonBox.rejected.connect(NewTagDialog.reject)
        self.pushCheckBox.toggled['bool'].connect(self.remoteComboBox.setEnabled)
        NewTagDialog.setTabOrder(self.nameEdit, self.pushCheckBox)
        NewTagDialog.setTabOrder(self.pushCheckBox, self.remoteComboBox)
        NewTagDialog.setTabOrder(self.remoteComboBox, self.forceCheckBox)

    def retranslateUi(self, NewTagDialog):
        NewTagDialog.setWindowTitle(_p("NewTagDialog", "New tag"))
        self.label.setText(_p("NewTagDialog", "&Name:"))
        self.nameEdit.setPlaceholderText(_p("NewTagDialog", "Enter tag name"))
        self.pushCheckBox.setText(_p("NewTagDialog", "&Push to:"))
        self.forceCheckBox.setText(_p("NewTagDialog", "Replace existing tag (locally only)"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import enum
import logging
from pathlib import Path

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class DropAction(enum.IntEnum):
    Deny = enum.auto()
    Open = enum.auto()
    Patch = enum.auto()
    Clone = enum.auto()
    Blame = enum.auto()


class DropZone(QWidget):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        font: QFont = self.font()
        font.setPointSizeF(font.pointSizeF() * 2)
        self.setFont(font)
        self.message = ""
        self.icon = "action-unavailable"

    def install(self, action: DropAction, data: str):
        if action == DropAction.Open:
            path = Path(data)
            self.message = _("Drop here to open repo {0}", tquoe(path.name))
            self.icon = "git-folder"
        elif action == DropAction.Clone:
            self.message = _("Drop here to clone {0}", tquoe(data))
            self.icon = "git-remote"
        elif action == DropAction.Patch:
            path = Path(data)
            self.message = _("Drop here to apply patch {0}", tquoe(path.name))
            self.icon = "git-stage-lines"
        elif action == DropAction.Blame:
            path = Path(data)
            self.message = _("Drop here to blame {0}", tquoe(path.name))
            self.icon = "git-blame"
        else:  # Deny
            self.message = data
            self.icon = "action-unavailable"

        self.setFixedSize(self.window().geometry().size())
        self.setVisible(True)

    def paintEvent(self, event: QPaintEvent):
        gray = QColor(128, 128, 128, 230)
        white = Qt.GlobalColor.white
        pad = 32
        iconSize = 16 * 5
        iconPad = iconSize // 4

        painter = QPainter(self)

        rect = QRectF(self.rect())

        painter.fillRect(rect, gray)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.setPen(QPen(white, 4, Qt.PenStyle.DotLine))

        iconRect = QRect(pad, pad, iconSize, iconSize)
        icon = stockIcon(self.icon, "gray=white")
        icon.paint(painter, iconRect)

        textOption = QTextOption()
        textOption.setAlignment(Qt.AlignmentFlag.AlignVCenter)
        textOption.setWrapMode(QTextOption.WrapMode.WordWrap)
        textRect = QRectF(0, pad, rect.width(), iconRect.height())
        textRect.adjust(iconRect.right() + iconPad, -iconPad, -pad, iconPad)
        painter.drawText(textRect, self.message, textOption)

        rect.adjust(pad, pad, -pad, -pad)
        rect.setTop(iconRect.bottom() + iconPad)
        painter.drawRoundedRect(rect, pad, pad)

        painter.end()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import os
import stat
from pathlib import Path

import mfusepy as fuse
from pygit2 import Blob, Commit, Object, Tree, Repository, Reference
from pygit2.enums import ObjectType

logger = logging.getLogger(__name__)


class TreeMount(fuse.Operations):
    use_ns = True  # squelch mfusepy warning

    @classmethod
    def run(cls, workdir: str, lookup: str, mountPoint: str):
        repo = Repository(workdir)
        try:
            gitObj: Reference | Object
            try:
                gitObj = repo[lookup]
            except ValueError:
                gitObj = repo.lookup_reference_dwim(lookup)
            commit = gitObj.peel(Commit)
            ops = cls(commit, mountPoint)
            fuse.FUSE(ops, mountPoint, foreground=True, ro=True)
        finally:
            repo.free()

    def __init__(self, commit: Commit, mountPoint: str) -> None:
        super().__init__()
        self.authorTime = commit.author.time * 1_000_000_000
        self.committerTime = commit.committer.time * 1_000_000_000
        self.tree = commit.tree
        self.mountPointPathObj = Path(mountPoint)  # to produce absolute paths in readlink

        try:
            self.defaultGid, self.defaultUid = os.getgid(), os.getuid()
        except AttributeError:  # Windows doesn't have these
            self.defaultGid, self.defaultUid = 0, 0

    def _resolve(self, path: str) -> Object:
        assert path.startswith("/")
        path = path.removeprefix("/")
        obj: Object = self.tree
        if path:
            for part in path.split("/"):
                tree = obj.peel(Tree)
                obj = tree[part]
        return obj

    @fuse.overrides(fuse.Operations)  # type: ignore[untyped-decorator]
    def getattr(self, path: str, fh: int) -> dict[str, int]:
        try:
            o = self._resolve(path)
        except KeyError:
            logger.warning(f"Could not resolve: {path}")
            return {}

        if o.type in (ObjectType.TREE, ObjectType.COMMIT):
            size = 0
            mode = stat.S_IFDIR | 0o111
        elif o.type == ObjectType.BLOB:
            assert isinstance(o, Blob)
            size = o.size
            mode = o.filemode
        else:
            logger.warning(f"Unsupported object type {o.type} for: {path}")
            return {}

        mode |= 0o444  # Make everything readable

        return {
            "st_birthtime": self.authorTime,  # Only relevant on macOS
            "st_ctime": self.committerTime,
            "st_atime": self.committerTime,
            "st_mtime": self.committerTime,
            "st_gid": self.defaultGid,
            "st_uid": self.defaultUid,
            "st_size": size,
            "st_mode": mode,
        }

    @fuse.overrides(fuse.Operations)  # type: ignore[untyped-decorator]
    def readdir(self, path: str, fh: int) -> fuse.ReadDirResult:
        o = self._resolve(path)
        if o.type == ObjectType.TREE:
            tree = o.peel(Tree)
            return [".", ".."] + [entry.name for entry in tree]
        elif o.type == ObjectType.COMMIT:
            # TODO: Submodules
            return [".", ".."]
        else:
            return []

    @fuse.overrides(fuse.Operations)  # type: ignore[untyped-decorator]
    def read(self, path: str, size: int, offset: int, fh: int) -> bytes:
        blob = self._resolve(path).peel(Blob)
        return blob.data[offset: offset+size]

    @fuse.overrides(fuse.Operations)  # type: ignore[untyped-decorator]
    def readlink(self, path: str) -> str:
        blobText = self._resolve(path).peel(Blob).data.decode("utf-8")
        absTarget = Path(path).parent / blobText
        absOnHost = self.mountPointPathObj / absTarget.relative_to("/")
        return str(absOnHost)


def main():
    import argparse
    parser = argparse.ArgumentParser(description="Mount Git tree with FUSE")
    parser.add_argument("repo")
    parser.add_argument("refish")
    parser.add_argument("mountpoint")
    args = parser.parse_args()
    TreeMount.run(args.repo, args.refish, args.mountpoint)


if __name__ == '__main__':
    main()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.statusform import StatusForm
from gitfourchette.qt import *
from gitfourchette.toolbox import QProcessConnection, QSignalBlockerContext


class ProcessDialog(QDialog):
    PopUpDelay = 300 if not WINDOWS else 600

    becameVisible = Signal()

    processConnection: QProcessConnection
    sentSigterm: bool

    def __init__(self, parent: QWidget):
        super().__init__(parent)

        self.processConnection = QProcessConnection(self)
        self.processConnection.processLost.connect(self.onProcessLost)

        self.sentSigterm = False

        # Set up UI
        layout = QVBoxLayout(self)
        self.statusForm = StatusForm(self)
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setFocusPolicy(Qt.FocusPolicy.ClickFocus)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.abortButton = self.buttonBox.addButton(QDialogButtonBox.StandardButton.Abort)
        layout.addWidget(self.statusForm)
        layout.addWidget(self.buttonBox)

        # Delay popup to avoid flashing when the process finishes fast enough.
        self.delayPopUp = QTimer(self)
        self.delayPopUp.timeout.connect(self.popUp)
        self.delayPopUp.setSingleShot(True)
        self.delayPopUp.setInterval(ProcessDialog.PopUpDelay)

        self.setMinimumWidth(self.fontMetrics().horizontalAdvance("W" * 40))
        self.setWindowFlags(Qt.WindowType.Dialog | Qt.WindowType.CustomizeWindowHint | Qt.WindowType.WindowTitleHint)  # hide close button
        self.setWindowModality(Qt.WindowModality.WindowModal)

        self.resize(self.width(), 1)

        # Remove keyboard focus from the abort button to prevent accidental triggering.
        # The user can still press the Escape key to abort.
        self.abortButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        self.statusForm.connectAbortButton(self.abortButton)

    def connectProcess(self, process: QProcess, title: str):
        # Forget existing process, if any, and keep track of this one.
        # Block signal 'processConnection.processLost' to avoid closing and
        # reopening the dialog when chaining processes quickly.
        with QSignalBlockerContext(self.processConnection):
            self.processConnection.track(process)

        self.setWindowTitle(title)

        if self.isVisible():
            # Dialog already shown, update UI now
            self.connectProcessInStatusForm()
        elif not self.delayPopUp.isActive():
            # Delay popup to avoid ugly flashing + unnecessary layout work
            # if the process finishes fast enough
            self.delayPopUp.start()

    def popUp(self):
        if not self.processConnection:
            return

        self.connectProcessInStatusForm()
        self.show()
        self.becameVisible.emit()

    def connectProcessInStatusForm(self):
        self.statusForm.connectProcess(self.processConnection.process)

    def onProcessLost(self):
        self.delayPopUp.stop()
        self.setVisible(False)

        # HACK: In offscreen unit tests, re-activate the previous window so we can assume
        # keyboard focus is still valid. Qt does this for us in non-offscreen mode.
        if APP_TESTMODE and OFFSCREEN:
            self.parentWidget().activateWindow()

    def close(self) -> bool:
        self.processConnection.stopTracking()
        return super().close()

    def reject(self):  # bound to abort button and ESC key
        self.statusForm.requestAbort()



# Form implementation generated from reading ui file 'aboutdialog.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        AboutDialog.setObjectName("AboutDialog")
        AboutDialog.setWindowModality(Qt.WindowModality.ApplicationModal)
        AboutDialog.resize(500, 500)
        self.gridLayout = QGridLayout(AboutDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.iconLabel = QLabel(parent=AboutDialog)
        self.iconLabel.setText("ICON")
        self.iconLabel.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.iconLabel.setObjectName("iconLabel")
        self.gridLayout.addWidget(self.iconLabel, 0, 0, 1, 1)
        self.tabWidget = QTabWidget(parent=AboutDialog)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout_2 = QVBoxLayout(self.tab)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.scrollArea = QScrollArea(parent=self.tab)
        self.scrollArea.setFrameShape(QFrame.Shape.NoFrame)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents_2 = QWidget()
        self.scrollAreaWidgetContents_2.setGeometry(QRect(0, 0, 482, 345))
        self.scrollAreaWidgetContents_2.setObjectName("scrollAreaWidgetContents_2")
        self.horizontalLayout = QHBoxLayout(self.scrollAreaWidgetContents_2)
        self.horizontalLayout.setContentsMargins(32, 16, 32, 16)
        self.horizontalLayout.setSpacing(16)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.aboutBlurb = QLabel(parent=self.scrollAreaWidgetContents_2)
        self.aboutBlurb.setText("ABOUT")
        self.aboutBlurb.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.aboutBlurb.setWordWrap(True)
        self.aboutBlurb.setOpenExternalLinks(True)
        self.aboutBlurb.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextSelectableByKeyboard|Qt.TextInteractionFlag.TextSelectableByMouse)
        self.aboutBlurb.setObjectName("aboutBlurb")
        self.horizontalLayout.addWidget(self.aboutBlurb)
        self.mugshot = QLabel(parent=self.scrollAreaWidgetContents_2)
        self.mugshot.setText("MUG")
        self.mugshot.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTop|Qt.AlignmentFlag.AlignTrailing)
        self.mugshot.setObjectName("mugshot")
        self.horizontalLayout.addWidget(self.mugshot)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents_2)
        self.verticalLayout_2.addWidget(self.scrollArea)
        self.tabWidget.addTab(self.tab, "")
        self.tab_4 = QWidget()
        self.tab_4.setObjectName("tab_4")
        self.verticalLayout_7 = QVBoxLayout(self.tab_4)
        self.verticalLayout_7.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.scrollArea_4 = QScrollArea(parent=self.tab_4)
        self.scrollArea_4.setFrameShape(QFrame.Shape.NoFrame)
        self.scrollArea_4.setWidgetResizable(True)
        self.scrollArea_4.setObjectName("scrollArea_4")
        self.scrollAreaWidgetContents_4 = QWidget()
        self.scrollAreaWidgetContents_4.setGeometry(QRect(0, 0, 117, 48))
        self.scrollAreaWidgetContents_4.setObjectName("scrollAreaWidgetContents_4")
        self.verticalLayout_6 = QVBoxLayout(self.scrollAreaWidgetContents_4)
        self.verticalLayout_6.setContentsMargins(32, 16, 32, 16)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.ackBlurb = QLabel(parent=self.scrollAreaWidgetContents_4)
        self.ackBlurb.setText("LICENSE")
        self.ackBlurb.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.ackBlurb.setWordWrap(True)
        self.ackBlurb.setOpenExternalLinks(True)
        self.ackBlurb.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByKeyboard|Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextBrowserInteraction|Qt.TextInteractionFlag.TextSelectableByKeyboard|Qt.TextInteractionFlag.TextSelectableByMouse)
        self.ackBlurb.setObjectName("ackBlurb")
        self.verticalLayout_6.addWidget(self.ackBlurb)
        self.scrollArea_4.setWidget(self.scrollAreaWidgetContents_4)
        self.verticalLayout_7.addWidget(self.scrollArea_4)
        self.tabWidget.addTab(self.tab_4, "")
        self.tab_3 = QWidget()
        self.tab_3.setObjectName("tab_3")
        self.verticalLayout = QVBoxLayout(self.tab_3)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.scrollArea_3 = QScrollArea(parent=self.tab_3)
        self.scrollArea_3.setFrameShape(QFrame.Shape.NoFrame)
        self.scrollArea_3.setWidgetResizable(True)
        self.scrollArea_3.setObjectName("scrollArea_3")
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 117, 48))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.verticalLayout_5 = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_5.setContentsMargins(32, 16, 32, 16)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.licenseBlurb = QLabel(parent=self.scrollAreaWidgetContents)
        self.licenseBlurb.setText("LICENSE")
        self.licenseBlurb.setAlignment(Qt.AlignmentFlag.AlignLeading|Qt.AlignmentFlag.AlignLeft|Qt.AlignmentFlag.AlignTop)
        self.licenseBlurb.setWordWrap(True)
        self.licenseBlurb.setOpenExternalLinks(True)
        self.licenseBlurb.setTextInteractionFlags(Qt.TextInteractionFlag.LinksAccessibleByKeyboard|Qt.TextInteractionFlag.LinksAccessibleByMouse|Qt.TextInteractionFlag.TextBrowserInteraction|Qt.TextInteractionFlag.TextSelectableByKeyboard|Qt.TextInteractionFlag.TextSelectableByMouse)
        self.licenseBlurb.setObjectName("licenseBlurb")
        self.verticalLayout_5.addWidget(self.licenseBlurb)
        self.scrollArea_3.setWidget(self.scrollAreaWidgetContents)
        self.verticalLayout.addWidget(self.scrollArea_3)
        self.tabWidget.addTab(self.tab_3, "")
        self.gridLayout.addWidget(self.tabWidget, 3, 0, 1, 4)
        self.buttonBox = QDialogButtonBox(parent=AboutDialog)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 4, 3, 1, 1)
        self.header = QLabel(parent=AboutDialog)
        self.header.setText("HEADER")
        self.header.setOpenExternalLinks(True)
        self.header.setObjectName("header")
        self.gridLayout.addWidget(self.header, 0, 2, 1, 2)
        spacerItem = QSpacerItem(8, 20, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 1, 1, 1)
        spacerItem1 = QSpacerItem(20, 16, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.gridLayout.addItem(spacerItem1, 2, 2, 1, 1)
        self.versionLabel = QLabel(parent=AboutDialog)
        self.versionLabel.setText("VERSION")
        self.versionLabel.setObjectName("versionLabel")
        self.gridLayout.addWidget(self.versionLabel, 1, 2, 1, 2)

        self.retranslateUi(AboutDialog)
        self.tabWidget.setCurrentIndex(0)
        self.buttonBox.accepted.connect(AboutDialog.close)

    def retranslateUi(self, AboutDialog):
        AboutDialog.setWindowTitle(_p("AboutDialog", "About {0}"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _p("AboutDialog", "About"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_4), _p("AboutDialog", "Acknowledgments"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_3), _p("AboutDialog", "License"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from pathlib import Path

from gitfourchette.gitdriver import GitDelta, GitStatus, GitDriver
from gitfourchette.localization import *
from gitfourchette.porcelain import *
from gitfourchette.tasks.repotask import AbortTask, RepoTask, TaskEffects
from gitfourchette.toolbox import *


def savePatch(task: RepoTask, patch: str, fileName="") -> RepoTask.Flow[str]:
    if not patch:
        raise AbortTask(_("Nothing to export. The patch is empty."), icon="information")

    # Sanitize filename
    for c in "?/\\*~<>|:":
        fileName = fileName.replace(c, "_")

    qfd = PersistentFileDialog.saveFile(task.parentWidget(), "SaveFile", task.name(), fileName)
    savePath = yield from task.flowFileDialog(qfd)

    yield from task.flowEnterWorkerThread()
    # Write patches verbatim: force UTF-8 (not the locale encoding, which fails
    # on non-cp1252 content on Windows) and keep LF line endings (CRLF
    # translation corrupts --binary base85 payloads and can break git apply).
    Path(savePath).write_text(patch, encoding="utf-8", newline="\n")

    if task.repo.is_in_workdir(savePath):
        task.epilog.effects |= TaskEffects.Workdir  # invalidate workdir if saved file to it

    task.epilog.status = _("Patch saved as: {0}", tquo(compactPath(savePath)))
    return savePath


class ExportCommitAsPatch(RepoTask):
    def flow(self, oid: Oid, fileName=""):
        commit = self.repo.peel_commit(oid)

        if not fileName:
            summary, _dummy = messageSummary(commit.message, elision="")
            summary = summary[:50].strip()
            fileName = f"{self.repo.repo_name()} - {shortHash(oid)} - {summary}.patch"

        diffAB = commit_diff_pair(commit)
        tokens = GitDriver.buildDiffCommand(diffAB)
        driver = yield from self.flowCallGit(*tokens)
        patch = driver.stdoutScrollback()

        yield from savePatch(self, patch, fileName)


class ExportStashAsPatch(ExportCommitAsPatch):
    def flow(self, oid: Oid, fileName=""):
        if not fileName:
            commit = self.repo.peel_commit(oid)
            message = _p("patch file name, please keep it short",
                         "stashed on {commit}",
                         commit=shortHash(commit.parent_ids[0]))
            summary = strip_stash_message(commit.message)[:50].strip()
            fileName = f"{self.repo.repo_name()} - {message} - {summary}.patch"

        yield from super().flow(oid, fileName)


class ExportABDiffAsPatch(RepoTask):
    def flow(self, diffAB: tuple[Oid, Oid]):
        fileName = f"{self.repo.repo_name()} - {shortHash(diffAB[0])}...{shortHash(diffAB[1])}.patch"

        tokens = GitDriver.buildDiffCommand(diffAB)
        driver = yield from self.flowCallGit(*tokens)
        patch = driver.stdoutScrollback()

        yield from savePatch(self, patch, fileName)


class ExportWorkdirAsPatch(RepoTask):
    def flow(self):
        patches = []

        # Diff the workdir to HEAD (except untracked files)
        tokens = GitDriver.buildDiffCommand(None)
        driver = yield from self.flowCallGit(*tokens, "HEAD")
        patches.append(driver.stdoutScrollback())

        # Diff untracked files.
        # This requires fresh workdir status in RepoModel, which we probably have already
        # because this task requires the user to click on the workdir first.
        if not self.repoModel.workdirStatusReady or self.repoModel.workdirStale:
            raise NotImplementedError("Export workdir requires fresh status")

        for delta in self.repoModel.workdirUnstagedDeltas:
            if delta.status == GitStatus.Untracked:  # Scan for untracked files
                tokens = GitDriver.buildDiffCommand(delta)
                driver = yield from self.flowCallGit(*tokens, autoFail=False)
                patches.append(driver.stdoutScrollback())

        # Compose the patch
        assert all(not patch or patch.endswith("\n") for patch in patches)
        patch = "".join(patches)

        # Compose filename
        message = _p("patch file name, please keep it short",
                     "uncommitted changes on {commit}", commit=shortHash(self.repo.head_commit_id))
        fileName = f"{self.repo.repo_name()} - {message}.patch"

        yield from savePatch(self, patch, fileName)


class ExportPatchCollection(RepoTask):
    def flow(self, deltas: list[GitDelta]):
        names = []
        patches = []

        for delta in deltas:
            # Get filename stem
            file = delta.old if delta.status == GitStatus.Deleted else delta.new
            name = Path(file.path).stem
            names.append(name)

            # Get patch (run 'git diff')
            tokens = GitDriver.buildDiffCommand(delta)
            driver = yield from self.flowCallGit(*tokens, autoFail=False)
            patch = driver.stdoutScrollback()
            patches.append(patch)

        # Compose patch and filename
        assert all(not patch or patch.endswith("\n") for patch in patches)
        composed = "".join(patches)
        fileName = ", ".join(names) + ".patch"

        yield from savePatch(self, composed, fileName)



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.ui_welcomewidget import Ui_WelcomeWidget
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class WelcomeWidget(QFrame):
    newRepo = Signal()
    openRepo = Signal()
    cloneRepo = Signal()

    def __init__(self, parent):
        super().__init__(parent)

        self.ui = Ui_WelcomeWidget()
        self.ui.setupUi(self)

        logoPixmap = QPixmap("assets:icons/gitfourchette")
        logoPixmap.setDevicePixelRatio(4)
        self.ui.logoLabel.setText(qAppName())
        self.ui.logoLabel.setPixmap(logoPixmap)

        defaultFont = self.ui.welcomeLabel.font()
        fs1 = int(defaultFont.pointSizeF() * 1.3)
        fs2 = int(defaultFont.pointSizeF() * 1.8)
        appText = f"<span style='font-weight: bold; color: #407cbf; font-size: {fs2}pt'>{qAppName()}</span>"
        welcomeText = self.ui.welcomeLabel.text()
        welcomeText = f"<html style='font-size: {fs1}pt;'>" + welcomeText.format(app=appText)
        self.ui.welcomeLabel.setText(welcomeText)

        self.ui.newRepoButton.clicked.connect(self.newRepo)
        self.ui.openRepoButton.clicked.connect(self.openRepo)
        self.ui.cloneRepoButton.clicked.connect(self.cloneRepo)
